# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import itertools
import operator
from contextlib import contextmanager
from numbers import Integral
from typing import List, Union

import numpy as np
import pandas as pd
from pandas.api.types import is_string_dtype
from pandas.api.extensions import ExtensionDtype
from pandas.core.dtypes.cast import find_common_type

from ..config import options
from ..core import Entity, ExecutableTuple
from ..core.context import Context
from ..lib.mmh3 import hash as mmh_hash
from ..tensor.utils import dictify_chunk_size, normalize_chunk_sizes
from ..typing import ChunkType, TileableType
from ..utils import (
    tokenize,
    sbytes,
    lazy_import,
    ModulePlaceholder,
    is_full_slice,
    parse_readable_size,
)

try:
    import pyarrow as pa
except ImportError:  # pragma: no cover
    pa = ModulePlaceholder("pyarrow")

cudf = lazy_import("cudf", rename="cudf")


def hash_index(index, size):
    def func(x, size):
        return mmh_hash(sbytes(x)) % size

    f = functools.partial(func, size=size)
    idx_to_grouped = index.groupby(index.map(f))
    return [idx_to_grouped.get(i, list()) for i in range(size)]


def hash_dataframe_on(df, on, size, level=None):
    if on is None:
        idx = df.index
        if level is not None:
            idx = idx.to_frame(False)[level]
        if cudf and isinstance(idx, cudf.Index):  # pragma: no cover
            idx = idx.to_pandas()
        hashed_label = pd.util.hash_pandas_object(idx, categorize=False)
    elif callable(on):
        # todo optimization can be added, if ``on`` is a numpy ufunc or sth can be vectorized
        hashed_label = pd.util.hash_pandas_object(df.index.map(on), categorize=False)
    else:
        if isinstance(on, list):
            to_concat = []
            for v in on:
                if isinstance(v, pd.Series):
                    to_concat.append(v)
                else:
                    to_concat.append(df[v])
            data = pd.concat(to_concat, axis=1)
        else:
            data = df[on]
        hashed_label = pd.util.hash_pandas_object(data, index=False, categorize=False)
    idx_to_grouped = pd.RangeIndex(0, len(hashed_label)).groupby(hashed_label % size)
    return [idx_to_grouped.get(i, pd.Index([])) for i in range(size)]


def hash_dtypes(dtypes, size):
    hashed_indexes = hash_index(dtypes.index, size)
    return [dtypes[index] for index in hashed_indexes]


def sort_dataframe_inplace(df, *axis):
    for ax in axis:
        df.sort_index(axis=ax, inplace=True)
    return df


@functools.lru_cache(1)
def _get_range_index_type():
    if cudf is not None:
        return pd.RangeIndex, cudf.RangeIndex
    else:
        return pd.RangeIndex


@functools.lru_cache(1)
def _get_multi_index_type():
    if cudf is not None:
        return pd.MultiIndex, cudf.MultiIndex
    else:
        return pd.MultiIndex


def _get_range_index_start(pd_range_index):
    try:
        return pd_range_index.start
    except AttributeError:  # pragma: no cover
        return pd_range_index._start


def _get_range_index_stop(pd_range_index):
    try:
        return pd_range_index.stop
    except AttributeError:  # pragma: no cover
        return pd_range_index._stop


def _get_range_index_step(pd_range_index):
    try:
        return pd_range_index.step
    except AttributeError:  # pragma: no cover
        pass
    try:  # pragma: no cover
        return pd_range_index._step
    except AttributeError:  # pragma: no cover
        return 1  # cudf does not support step arg


def is_pd_range_empty(pd_range_index):
    start, stop, step = (
        _get_range_index_start(pd_range_index),
        _get_range_index_stop(pd_range_index),
        _get_range_index_step(pd_range_index),
    )
    return (start >= stop and step >= 0) or (start <= stop and step < 0)


def decide_dataframe_chunk_sizes(shape, chunk_size, memory_usage):
    """
    Decide how a given DataFrame can be split into chunk.

    :param shape: DataFrame's shape
    :param chunk_size: if dict provided, it's dimension id to chunk size;
                       if provided, it's the chunk size for each dimension.
    :param memory_usage: pandas Series in which each column's memory usage
    :type memory_usage: pandas.Series
    :return: the calculated chunk size for each dimension
    :rtype: tuple
    """
    from ..config import options

    chunk_size = dictify_chunk_size(shape, chunk_size)
    average_memory_usage = memory_usage / shape[0]

    nleft = len(shape) - len(chunk_size)
    if nleft < 0:
        raise ValueError("chunks have more than two dimensions")
    if nleft == 0:
        return normalize_chunk_sizes(
            shape, tuple(chunk_size[j] for j in range(len(shape)))
        )

    max_chunk_size = options.chunk_store_limit

    # for the row side, along axis 0
    if 0 not in chunk_size:
        row_chunk_size = []
        row_left_size = shape[0]
    else:
        row_chunk_size = normalize_chunk_sizes((shape[0],), (chunk_size[0],))[0]
        row_left_size = -1
    # for the column side, along axis 1
    if 1 not in chunk_size:
        col_chunk_size = []
        col_chunk_store = []
        col_left_size = shape[1]
    else:
        col_chunk_size = normalize_chunk_sizes((shape[1],), (chunk_size[1],))[0]
        acc = [0] + np.cumsum(col_chunk_size).tolist()
        col_chunk_store = [
            average_memory_usage[acc[i] : acc[i + 1]].sum()
            for i in range(len(col_chunk_size))
        ]
        col_left_size = -1

    while True:
        nbytes_occupied = np.prod(
            [max(it) for it in (row_chunk_size, col_chunk_store) if it]
        )
        dim_size = np.maximum(
            int(np.power(max_chunk_size / nbytes_occupied, 1 / float(nleft))), 1
        )

        if col_left_size == 0 and not col_chunk_size:
            col_chunk_size.append(0)

        if row_left_size == 0 and not row_chunk_size:
            row_chunk_size.append(0)

        # check col first
        if col_left_size > 0:
            cs = min(col_left_size, dim_size)
            col_chunk_size.append(cs)
            start = int(np.sum(col_chunk_size[:-1]))
            col_chunk_store.append(average_memory_usage.iloc[start : start + cs].sum())
            col_left_size -= cs
        if row_left_size > 0:
            if col_chunk_store:
                max_col_chunk_store = max(col_chunk_store)
                cs = min(row_left_size, int(max_chunk_size / max_col_chunk_store))
            else:
                cs = row_left_size
            row_chunk_size.append(cs)
            row_left_size -= cs

        if col_left_size <= 0 and row_left_size <= 0:
            break

    return tuple(row_chunk_size), tuple(col_chunk_size)


def decide_series_chunk_size(shape, chunk_size, memory_usage):
    from ..config import options

    chunk_size = dictify_chunk_size(shape, chunk_size)
    average_memory_usage = memory_usage / shape[0] if shape[0] != 0 else memory_usage

    if len(chunk_size) == len(shape):
        return normalize_chunk_sizes(shape, chunk_size[0])

    if all(s == 0 for s in shape):
        # skip when shape is 0
        return tuple((s,) for s in shape)

    max_chunk_size = options.chunk_store_limit
    series_chunk_size = max_chunk_size / average_memory_usage
    return normalize_chunk_sizes(shape, int(series_chunk_size))


def parse_index(index_value, *args, store_data=False, key=None):
    from .core import IndexValue

    def _extract_property(index, tp, ret_data):
        kw = {
            "_min_val": _get_index_min(index),
            "_max_val": _get_index_max(index),
            "_min_val_close": True,
            "_max_val_close": True,
            "_key": key or _tokenize_index(index, *args),
        }
        if ret_data:
            kw["_data"] = index.values
        for field in tp._FIELDS:
            if field in kw or field == "_data":
                continue
            val = getattr(index, field.lstrip("_"), None)
            if val is not None:
                kw[field] = val
        return kw

    def _tokenize_index(index, *token_objects):
        if not index.empty:
            return tokenize(index)
        else:
            return tokenize(index, *token_objects)

    def _get_index_min(index):
        try:
            return index.min()
        except (ValueError, AttributeError):
            if isinstance(index, pd.IntervalIndex):
                return None
            raise
        except TypeError:
            return None

    def _get_index_max(index):
        try:
            return index.max()
        except (ValueError, AttributeError):
            if isinstance(index, pd.IntervalIndex):
                return None
            raise
        except TypeError:
            return None

    def _serialize_index(index):
        tp = getattr(IndexValue, type(index).__name__)
        properties = _extract_property(index, tp, store_data)
        properties["_name"] = index.name
        return tp(**properties)

    def _serialize_range_index(index):
        if is_pd_range_empty(index):
            properties = {
                "_is_monotonic_increasing": True,
                "_is_monotonic_decreasing": False,
                "_is_unique": True,
                "_min_val": _get_index_min(index),
                "_max_val": _get_index_max(index),
                "_min_val_close": True,
                "_max_val_close": False,
                "_key": key or _tokenize_index(index, *args),
                "_name": index.name,
                "_dtype": index.dtype,
            }
        else:
            properties = _extract_property(index, IndexValue.RangeIndex, False)
        return IndexValue.RangeIndex(
            _slice=slice(
                _get_range_index_start(index),
                _get_range_index_stop(index),
                _get_range_index_step(index),
            ),
            **properties,
        )

    def _serialize_multi_index(index):
        kw = _extract_property(index, IndexValue.MultiIndex, store_data)
        kw["_sortorder"] = index.sortorder
        kw["_dtypes"] = [lev.dtype for lev in index.levels]
        return IndexValue.MultiIndex(**kw)

    if index_value is None:
        return IndexValue(
            _index_value=IndexValue.Index(
                _is_monotonic_increasing=False,
                _is_monotonic_decreasing=False,
                _is_unique=False,
                _min_val=None,
                _max_val=None,
                _min_val_close=True,
                _max_val_close=True,
                _key=key or tokenize(*args),
            )
        )
    if hasattr(index_value, "to_pandas"):  # pragma: no cover
        # convert cudf.Index to pandas
        index_value = index_value.to_pandas()

    if isinstance(index_value, _get_range_index_type()):
        return IndexValue(_index_value=_serialize_range_index(index_value))
    elif isinstance(index_value, _get_multi_index_type()):
        return IndexValue(_index_value=_serialize_multi_index(index_value))
    else:
        return IndexValue(_index_value=_serialize_index(index_value))


def gen_unknown_index_value(index_value, *args):
    pd_index = index_value.to_pandas()
    if isinstance(pd_index, pd.RangeIndex):
        return parse_index(pd.RangeIndex(-1), *args)
    elif not isinstance(pd_index, pd.MultiIndex):
        return parse_index(pd.Index([], dtype=pd_index.dtype), *args)
    else:
        i = pd.MultiIndex.from_arrays(
            [c[:0] for c in pd_index.levels], names=pd_index.names
        )
        return parse_index(i, *args)


def split_monotonic_index_min_max(
    left_min_max, left_increase, right_min_max, right_increase
):
    """
    Split the original two min_max into new min_max. Each min_max should be a list
    in which each item should be a 4-tuple indicates that this chunk's min value,
    whether the min value is close, the max value, and whether the max value is close.
    The return value would be a nested list, each item is a list
    indicates that how this chunk should be split into.

    :param left_min_max: the left min_max
    :param left_increase: if the original data of left is increased
    :param right_min_max: the right min_max
    :param right_increase: if the original data of right is increased
    :return: nested list in which each item indicates how min_max is split

    >>> left_min_max = [(0, True, 3, True), (4, True, 8, True), (12, True, 18, True),
    ...                 (20, True, 22, True)]
    >>> right_min_max = [(2, True, 6, True), (7, True, 9, True), (10, True, 14, True),
    ...                  (18, True, 19, True)]
    >>> l, r = split_monotonic_index_min_max(left_min_max, True, right_min_max, True)
    >>> l
    [[(0, True, 2, False), (2, True, 3, True)], [(3, False, 4, False), (4, True, 6, True), (6, False, 7, False),
    (7, True, 8, True)], [(8, False, 9, True), (10, True, 12, False), (12, True, 14, True), (14, False, 18, False),
    (18, True, 18, True)], [(18, False, 19, True), [20, True, 22, True]]]
    >>> r
    [[(0, True, 2, False), (2, True, 3, True), (3, False, 4, False), (4, True, 6, True)],
    [(6, False, 7, False), (7, True, 8, True), (8, False, 9, True)], [(10, True, 12, False), (12, True, 14, True)],
    [(14, False, 18, False), (18, True, 18, True), (18, False, 19, True), [20, True, 22, True]]]
    """
    left_idx_to_min_max = [[] for _ in left_min_max]
    right_idx_to_min_max = [[] for _ in right_min_max]
    left_curr_min_max = list(left_min_max[0])
    right_curr_min_max = list(right_min_max[0])
    left_curr_idx = right_curr_idx = 0
    left_terminate = right_terminate = False

    while not left_terminate or not right_terminate:
        if left_terminate:
            left_idx_to_min_max[left_curr_idx].append(tuple(right_curr_min_max))
            right_idx_to_min_max[right_curr_idx].append(tuple(right_curr_min_max))
            if right_curr_idx + 1 >= len(right_min_max):
                right_terminate = True
            else:
                right_curr_idx += 1
                right_curr_min_max = list(right_min_max[right_curr_idx])
        elif right_terminate:
            right_idx_to_min_max[right_curr_idx].append(tuple(left_curr_min_max))
            left_idx_to_min_max[left_curr_idx].append(tuple(left_curr_min_max))
            if left_curr_idx + 1 >= len(left_min_max):
                left_terminate = True
            else:
                left_curr_idx += 1
                left_curr_min_max = list(left_min_max[left_curr_idx])
        elif left_curr_min_max[0] < right_curr_min_max[0]:
            # left min < right min
            right_min = [right_curr_min_max[0], not right_curr_min_max[1]]
            max_val = min(left_curr_min_max[2:], right_min)
            assert len(max_val) == 2
            min_max = (
                left_curr_min_max[0],
                left_curr_min_max[1],
                max_val[0],
                max_val[1],
            )
            left_idx_to_min_max[left_curr_idx].append(min_max)
            right_idx_to_min_max[right_curr_idx].append(min_max)
            if left_curr_min_max[2:] == max_val:
                # left max < right min
                if left_curr_idx + 1 >= len(left_min_max):
                    left_terminate = True
                else:
                    left_curr_idx += 1
                    left_curr_min_max = list(left_min_max[left_curr_idx])
            else:
                # from left min(left min close) to right min(exclude right min close)
                left_curr_min_max[:2] = right_curr_min_max[:2]
        elif left_curr_min_max[0] > right_curr_min_max[0]:
            # left min > right min
            left_min = [left_curr_min_max[0], not left_curr_min_max[1]]
            max_val = min(right_curr_min_max[2:], left_min)
            min_max = (
                right_curr_min_max[0],
                right_curr_min_max[1],
                max_val[0],
                max_val[1],
            )
            left_idx_to_min_max[left_curr_idx].append(min_max)
            right_idx_to_min_max[right_curr_idx].append(min_max)
            if right_curr_min_max[2:] == max_val:
                # right max < left min
                if right_curr_idx + 1 >= len(right_min_max):
                    right_terminate = True
                else:
                    right_curr_idx += 1
                    right_curr_min_max = list(right_min_max[right_curr_idx])
            else:
                # from left min(left min close) to right min(exclude right min close)
                right_curr_min_max[:2] = left_curr_min_max[:2]
        else:
            # left min == right min
            max_val = min(left_curr_min_max[2:], right_curr_min_max[2:])
            assert len(max_val) == 2
            min_max = (
                left_curr_min_max[0],
                left_curr_min_max[1],
                max_val[0],
                max_val[1],
            )
            left_idx_to_min_max[left_curr_idx].append(min_max)
            right_idx_to_min_max[right_curr_idx].append(min_max)
            if max_val == left_curr_min_max[2:]:
                if left_curr_idx + 1 >= len(left_min_max):
                    left_terminate = True
                else:
                    left_curr_idx += 1
                    left_curr_min_max = list(left_min_max[left_curr_idx])
            else:
                left_curr_min_max[:2] = max_val[0], not max_val[1]
            if max_val == right_curr_min_max[2:]:
                if right_curr_idx + 1 >= len(right_min_max):
                    right_terminate = True
                else:
                    right_curr_idx += 1
                    right_curr_min_max = list(right_min_max[right_curr_idx])
            else:
                right_curr_min_max[:2] = max_val[0], not max_val[1]

    if left_increase is False:
        left_idx_to_min_max = list(reversed(left_idx_to_min_max))
    if right_increase is False:
        right_idx_to_min_max = list(reversed(right_idx_to_min_max))

    return left_idx_to_min_max, right_idx_to_min_max


def build_split_idx_to_origin_idx(splits, increase=True):
    # splits' len is equal to the original chunk size on a specified axis,
    # splits is sth like [[(0, True, 2, True), (2, False, 3, True)]]
    # which means there is one input chunk, and will be split into 2 out chunks
    # in this function, we want to build a new dict from the out chunk index to
    # the original chunk index and the inner position, like {0: (0, 0), 1: (0, 1)}
    if increase is False:
        splits = list(reversed(splits))
    out_idx = itertools.count(0)
    res = dict()
    for origin_idx, _ in enumerate(splits):
        for pos in range(len(splits[origin_idx])):
            if increase is False:
                o_idx = len(splits) - origin_idx - 1
            else:
                o_idx = origin_idx
            res[next(out_idx)] = o_idx, pos
    return res


def _generate_value(dtype, fill_value):
    # special handle for datetime64 and timedelta64
    dispatch = {
        np.datetime64: pd.Timestamp,
        np.timedelta64: pd.Timedelta,
        pd.CategoricalDtype.type: lambda x: pd.CategoricalDtype([x]),
        # for object, we do not know the actual dtype,
        # just convert to str for common usage
        np.object_: lambda x: str(fill_value),
    }
    # otherwise, just use dtype.type itself to convert
    convert = dispatch.get(dtype.type, dtype.type)
    return convert(fill_value)


def build_empty_df(dtypes, index=None):
    columns = dtypes.index
    length = len(index) if index is not None else 0
    record = [[_generate_value(dtype, 1) for dtype in dtypes]] * max(1, length)

    # duplicate column may exist,
    # so use RangeIndex first
    df = pd.DataFrame(record, columns=range(len(dtypes)), index=index)
    for i, dtype in enumerate(dtypes):
        s = df.iloc[:, i]
        if not pd.api.types.is_dtype_equal(s.dtype, dtype):
            df.iloc[:, i] = s.astype(dtype)

    df.columns = columns
    return df[:length] if len(df) > length else df


def build_df(df_obj, fill_value=1, size=1, ensure_string=False):
    dfs = []
    if not isinstance(size, (list, tuple)):
        sizes = [size]
    else:
        sizes = size

    if not isinstance(fill_value, (list, tuple)):
        fill_values = [fill_value]
    else:
        fill_values = fill_value

    for size, fill_value in zip(sizes, fill_values):
        dtypes = df_obj.dtypes
        record = [[_generate_value(dtype, fill_value) for dtype in dtypes]] * size
        df = pd.DataFrame(record)
        df.columns = dtypes.index

        if len(record) != 0:  # columns is empty in some cases
            target_index = df_obj.index_value.to_pandas()
            if isinstance(target_index, pd.MultiIndex):
                index_val = tuple(
                    _generate_value(level.dtype, fill_value)
                    for level in target_index.levels
                )
                df.index = pd.MultiIndex.from_tuples(
                    [index_val] * size, names=target_index.names
                )
            else:
                index_val = _generate_value(target_index.dtype, fill_value)
                df.index = pd.Index([index_val] * size, name=target_index.name)

        # make sure dtypes correct
        for i, dtype in enumerate(dtypes):
            s = df.iloc[:, i]
            if not pd.api.types.is_dtype_equal(s.dtype, dtype):
                df.iloc[:, i] = s.astype(dtype)
        dfs.append(df)
    if len(dfs) == 1:
        ret_df = dfs[0]
    else:
        ret_df = pd.concat(dfs)

    if ensure_string:
        obj_dtypes = df_obj.dtypes[df_obj.dtypes == np.dtype("O")]
        ret_df[obj_dtypes.index] = ret_df[obj_dtypes.index].radd("O")
    return ret_df


def build_empty_series(dtype, index=None, name=None):
    length = len(index) if index is not None else 0
    return pd.Series(
        [_generate_value(dtype, 1) for _ in range(length)],
        dtype=dtype,
        index=index,
        name=name,
    )


def build_series(
    series_obj=None,
    fill_value=1,
    size=1,
    name=None,
    ensure_string=False,
    dtype=None,
    index=None,
):
    seriess = []
    if not isinstance(size, (list, tuple)):
        sizes = [size]
    else:
        sizes = size

    if not isinstance(fill_value, (list, tuple)):
        fill_values = [fill_value]
    else:
        fill_values = fill_value

    if series_obj is not None:
        dtype = series_obj.dtype
        try:
            series_index = series_obj.index_value.to_pandas()[:0]
        except AttributeError:
            series_index = series_obj.index[:0]
    else:
        series_index = index[:0] if index is not None else None

    for size, fill_value in zip(sizes, fill_values):
        empty_series = build_empty_series(dtype, name=name, index=series_index)
        record = _generate_value(dtype, fill_value)
        if isinstance(empty_series.index, pd.MultiIndex):
            index = tuple(
                _generate_value(level.dtype, fill_value)
                for level in empty_series.index.levels
            )
            empty_series = empty_series.reindex(
                index=pd.MultiIndex.from_tuples([index], names=empty_series.index.names)
            )
            empty_series.iloc[0] = record
        else:
            if isinstance(empty_series.index.dtype, pd.CategoricalDtype):
                index = None
            else:
                index = _generate_value(empty_series.index.dtype, fill_value)
            empty_series.loc[index] = record

        empty_series = pd.concat([empty_series] * size)
        # make sure dtype correct for MultiIndex
        empty_series = empty_series.astype(dtype, copy=False)
        seriess.append(empty_series)

    if len(seriess) == 1:
        ret_series = seriess[0]
    else:
        ret_series = pd.concat(seriess)

    if ensure_string and dtype == np.dtype("O"):
        ret_series = ret_series.radd("O")
    return ret_series


def concat_index_value(index_values, store_data=False):
    if not isinstance(index_values, (list, tuple)):
        index_values = [index_values]
    result = index_values[0]
    if not isinstance(result, pd.Index):
        result = result.to_pandas()
    for index_value in index_values[1:]:
        if isinstance(index_value, pd.Index):
            result = result.append(index_value)
        else:
            result = result.append(index_value.to_pandas())
    return parse_index(result, store_data=store_data)


def build_concatenated_rows_frame(df):
    from ..core import OutputType
    from .merge.concat import DataFrameConcat

    # When the df isn't split along the column axis, return the df directly.
    if df.chunk_shape[1] == 1:
        return df

    columns = concat_index_value(
        [df.cix[0, idx].columns_value for idx in range(df.chunk_shape[1])],
        store_data=True,
    )
    columns_size = columns.to_pandas().size

    out_chunks = []
    for idx in range(df.chunk_shape[0]):
        out_chunk = DataFrameConcat(
            axis=1, output_types=[OutputType.dataframe]
        ).new_chunk(
            [df.cix[idx, k] for k in range(df.chunk_shape[1])],
            index=(idx, 0),
            shape=(df.cix[idx, 0].shape[0], columns_size),
            dtypes=df.dtypes,
            index_value=df.cix[idx, 0].index_value,
            columns_value=columns,
        )
        out_chunks.append(out_chunk)

    return DataFrameConcat(axis=1, output_types=[OutputType.dataframe]).new_dataframe(
        [df],
        chunks=out_chunks,
        nsplits=(tuple(chunk.shape[0] for chunk in out_chunks), (df.shape[1],)),
        shape=df.shape,
        dtypes=df.dtypes,
        index_value=df.index_value,
        columns_value=df.columns_value,
    )


def _filter_range_index(pd_range_index, min_val, min_val_close, max_val, max_val_close):
    if is_pd_range_empty(pd_range_index):
        return pd_range_index

    raw_min, raw_max, step = (
        pd_range_index.min(),
        pd_range_index.max(),
        _get_range_index_step(pd_range_index),
    )

    # seek min range
    greater_func = operator.gt if min_val_close else operator.ge
    actual_min = raw_min
    while greater_func(min_val, actual_min):
        actual_min += abs(step)
    if step < 0:
        actual_min += step  # on the right side

    # seek max range
    less_func = operator.lt if max_val_close else operator.le
    actual_max = raw_max
    while less_func(max_val, actual_max):
        actual_max -= abs(step)
    if step > 0:
        actual_max += step  # on the right side

    if step > 0:
        return pd.RangeIndex(actual_min, actual_max, step)
    return pd.RangeIndex(actual_max, actual_min, step)


def infer_index_value(left_index_value, right_index_value):
    from .core import IndexValue

    if isinstance(left_index_value.value, IndexValue.RangeIndex) and isinstance(
        right_index_value.value, IndexValue.RangeIndex
    ):
        if left_index_value.value.slice == right_index_value.value.slice:
            return left_index_value
        return parse_index(
            pd.Index([], dtype=np.int64), left_index_value, right_index_value
        )

    # when left index and right index is identical, and both of them are elements unique,
    # we can infer that the out index should be identical also
    if (
        left_index_value.is_unique
        and right_index_value.is_unique
        and left_index_value.key == right_index_value.key
    ):
        return left_index_value

    left_index = left_index_value.to_pandas()
    right_index = right_index_value.to_pandas()
    out_index = pd.Index(
        [], dtype=find_common_type([left_index.dtype, right_index.dtype])
    )
    return parse_index(out_index, left_index_value, right_index_value)


def filter_index_value(index_value, min_max, store_data=False):
    from .core import IndexValue

    min_val, min_val_close, max_val, max_val_close = min_max

    pd_index = index_value.to_pandas()

    if isinstance(index_value.value, IndexValue.RangeIndex):
        pd_filtered_index = _filter_range_index(
            pd_index, min_val, min_val_close, max_val, max_val_close
        )
        return parse_index(pd_filtered_index, store_data=store_data)

    if min_val_close:
        f = pd_index >= min_val
    else:
        f = pd_index > min_val
    if max_val_close:
        f = f & (pd_index <= max_val)
    else:
        f = f & (pd_index < max_val)

    return parse_index(pd_index[f], store_data=store_data)


def indexing_index_value(index_value, indexes, store_data=False, rechunk=False):
    pd_index = index_value.to_pandas()
    # when rechunk is True, the output index shall be treated
    # different from the input one
    if not rechunk and isinstance(indexes, slice) and is_full_slice(indexes):
        return index_value
    elif not index_value.has_value():
        new_index_value = parse_index(pd_index, indexes, store_data=store_data)
        new_index_value._index_value._min_val = index_value.min_val
        new_index_value._index_value._min_val_close = index_value.min_val_close
        new_index_value._index_value._max_val = index_value.max_val
        new_index_value._index_value._max_val_close = index_value.max_val_close
        return new_index_value
    else:
        if isinstance(indexes, Integral):
            return parse_index(pd_index[[indexes]], store_data=store_data)
        elif isinstance(indexes, Entity):
            if isinstance(pd_index, pd.RangeIndex):
                return parse_index(
                    pd.RangeIndex(-1), indexes, index_value, store_data=False
                )
            else:
                return parse_index(
                    type(pd_index)([]), indexes, index_value, store_data=False
                )
        if isinstance(indexes, tuple):
            return parse_index(pd_index[list(indexes)], store_data=store_data)
        else:
            return parse_index(pd_index[indexes], store_data=store_data)


def merge_index_value(to_merge_index_values: dict, store_data: bool = False):
    """
    Merge index value according to their chunk index.

    Parameters
    ----------
    to_merge_index_values : dict
        index to index_value
    store_data : bool
        store data in index_value

    Returns
    -------
    merged_index_value
    """

    pd_index = None
    min_val, min_val_close, max_val, max_val_close = None, None, None, None
    for _, chunk_index_value in sorted(to_merge_index_values.items()):
        if pd_index is None:
            pd_index = chunk_index_value.to_pandas()
            min_val, min_val_close, max_val, max_val_close = (
                chunk_index_value.min_val,
                chunk_index_value.min_val_close,
                chunk_index_value.max_val,
                chunk_index_value.max_val_close,
            )
        else:
            cur_pd_index = chunk_index_value.to_pandas()
            if store_data or (
                isinstance(pd_index, pd.RangeIndex)
                and isinstance(cur_pd_index, pd.RangeIndex)
                and cur_pd_index.step == pd_index.step
                and cur_pd_index.start == pd_index.stop
            ):
                # range index that is continuous
                pd_index = pd_index.append(cur_pd_index)
            else:
                pd_index = pd.Index([], dtype=pd_index.dtype)
            if chunk_index_value.min_val is not None:
                try:
                    if min_val is None or min_val > chunk_index_value.min_val:
                        min_val = chunk_index_value.min_val
                        min_val_close = chunk_index_value.min_val_close
                except TypeError:
                    # min_value has different types that cannot compare
                    # just stop compare
                    continue
            if chunk_index_value.max_val is not None:
                if max_val is None or max_val < chunk_index_value.max_val:
                    max_val = chunk_index_value.max_val
                    max_val_close = chunk_index_value.max_val_close

    index_value = parse_index(pd_index, store_data=store_data)
    if not index_value.has_value():
        index_value._index_value._min_val = min_val
        index_value._index_value._min_val_close = min_val_close
        index_value._index_value._max_val = max_val
        index_value._index_value._max_val_close = max_val_close
    return index_value


def infer_dtypes(left_dtypes, right_dtypes, operator):
    left = build_empty_df(left_dtypes)
    right = build_empty_df(right_dtypes)
    return operator(left, right).dtypes


@functools.lru_cache(100)
def infer_dtype(left_dtype, right_dtype, operator):
    left = build_empty_series(left_dtype)
    right = build_empty_series(right_dtype)
    return operator(left, right).dtype


def filter_dtypes(dtypes, column_min_max):
    left_filter = operator.ge if column_min_max[1] else operator.gt
    left = left_filter(dtypes.index, column_min_max[0])
    right_filter = operator.le if column_min_max[3] else operator.lt
    right = right_filter(dtypes.index, column_min_max[2])
    return dtypes[left & right]


def in_range_index(i, pd_range_index):
    """
    Check whether the input `i` is within `pd_range_index` which is a pd.RangeIndex.
    """
    start, stop, step = (
        _get_range_index_start(pd_range_index),
        _get_range_index_stop(pd_range_index),
        _get_range_index_step(pd_range_index),
    )
    if step > 0 and start <= i < stop and (i - start) % step == 0:
        return True
    if step < 0 and start >= i > stop and (start - i) % step == 0:
        return True
    return False


def wrap_notimplemented_exception(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except NotImplementedError:
            return NotImplemented

    return wrapper


def validate_axis(axis, tileable=None):
    if axis == "index":
        axis = 0
    elif axis == "columns":
        axis = 1

    illegal = False
    try:
        axis = operator.index(axis)
        if axis < 0 or (tileable is not None and axis >= tileable.ndim):
            illegal = True
    except TypeError:
        illegal = True

    if illegal:
        raise ValueError(f"No axis named {axis} for object type {type(tileable)}")
    return axis


def validate_axis_style_args(
    data, args, kwargs, arg_name, method_name
):  # pragma: no cover
    """Argument handler for mixed index, columns / axis functions

    In an attempt to handle both `.method(index, columns)`, and
    `.method(arg, axis=.)`, we have to do some bad things to argument
    parsing. This translates all arguments to `{index=., columns=.}` style.

    Parameters
    ----------
    data : DataFrame
    args : tuple
        All positional arguments from the user
    kwargs : dict
        All keyword arguments from the user
    arg_name, method_name : str
        Used for better error messages

    Returns
    -------
    kwargs : dict
        A dictionary of keyword arguments. Doesn't modify ``kwargs``
        inplace, so update them with the return value here.
    """
    out = {}
    # Goal: fill 'out' with index/columns-style arguments
    # like out = {'index': foo, 'columns': bar}

    # Start by validating for consistency
    axes_names = ["index"] if data.ndim == 1 else ["index", "columns"]
    if "axis" in kwargs and any(x in kwargs for x in axes_names):
        msg = "Cannot specify both 'axis' and any of 'index' or 'columns'."
        raise TypeError(msg)

    # First fill with explicit values provided by the user...
    if arg_name in kwargs:
        if args:
            msg = f"{method_name} got multiple values for argument '{arg_name}'"
            raise TypeError(msg)

        axis = axes_names[validate_axis(kwargs.get("axis", 0), data)]
        out[axis] = kwargs[arg_name]

    # More user-provided arguments, now from kwargs
    for k, v in kwargs.items():
        try:
            ax = axes_names[validate_axis(k, data)]
        except ValueError:
            pass
        else:
            out[ax] = v

    # All user-provided kwargs have been handled now.
    # Now we supplement with positional arguments, emitting warnings
    # when there's ambiguity and raising when there's conflicts

    if len(args) == 0:
        pass  # It's up to the function to decide if this is valid
    elif len(args) == 1:
        axis = axes_names[validate_axis(kwargs.get("axis", 0), data)]
        out[axis] = args[0]
    elif len(args) == 2:
        if "axis" in kwargs:
            # Unambiguously wrong
            msg = "Cannot specify both 'axis' and any of 'index' or 'columns'"
            raise TypeError(msg)

        msg = (
            "Interpreting call\n\t'.{method_name}(a, b)' as "
            "\n\t'.{method_name}(index=a, columns=b)'.\nUse named "
            "arguments to remove any ambiguity."
        )
        raise TypeError(msg.format(method_name=method_name))
    else:
        msg = f"Cannot specify all of '{arg_name}', 'index', 'columns'."
        raise TypeError(msg)
    return out


def validate_output_types(**kwargs):
    from ..core import OutputType

    output_type = kwargs.pop("object_type", None) or kwargs.pop("output_type", None)
    output_types = kwargs.pop("output_types", None) or (
        [output_type] if output_type is not None else None
    )
    return (
        [
            getattr(OutputType, v.lower()) if isinstance(v, str) else v
            for v in output_types
        ]
        if output_types
        else None
    )


def standardize_range_index(chunks: List[ChunkType], axis: int = 0):
    from .base.standardize_range_index import ChunkStandardizeRangeIndex

    row_chunks = dict(
        (k, next(v)) for k, v in itertools.groupby(chunks, key=lambda x: x.index[axis])
    )
    row_chunks = [row_chunks[i] for i in range(len(row_chunks))]

    out_chunks = []
    for c in chunks:
        prev_chunks = row_chunks[: c.index[axis]]
        op = ChunkStandardizeRangeIndex(
            prev_shapes=[p.shape for p in prev_chunks], axis=axis
        )
        op.output_types = c.op.output_types
        params = c.params.copy()
        start_pos = sum(p.shape[axis] for p in prev_chunks)
        end_pos = start_pos + c.shape[axis]
        index = pd.RangeIndex(start_pos, end_pos)
        if axis == 0:
            params["index_value"] = parse_index(index)
        else:
            dtypes = params["dtypes"]
            dtypes.index = index
            params["dtypes"] = dtypes
            params["columns_value"] = parse_index(dtypes.index, store_data=True)
        out_chunks.append(op.new_chunk([c], kws=[params]))

    return out_chunks


def fetch_corner_data(df_or_series, session=None) -> pd.DataFrame:
    """
    Fetch corner DataFrame or Series for repr usage.

    :param df_or_series: DataFrame or Series
    :return: corner DataFrame
    """
    from .indexing.iloc import iloc

    max_rows = pd.get_option("display.max_rows")
    try:
        min_rows = pd.get_option("display.min_rows")
        min_rows = min(min_rows, max_rows)
    except KeyError:  # pragma: no cover
        # display.min_rows is introduced in pandas 0.25
        min_rows = max_rows

    index_size = None
    if (
        df_or_series.shape[0] > max_rows
        and df_or_series.shape[0] > min_rows // 2 * 2 + 2
    ):
        # for pandas, greater than max_rows
        # will display min_rows
        # thus we fetch min_rows + 2 lines
        index_size = min_rows // 2 + 1

    if index_size is None:
        return df_or_series._fetch(session=session)
    else:
        head = iloc(df_or_series)[:index_size]
        tail = iloc(df_or_series)[-index_size:]
        head_data, tail_data = ExecutableTuple([head, tail]).fetch(session=session)
        xdf = cudf if head.op.is_gpu() else pd
        return xdf.concat([head_data, tail_data], axis="index")


class ReprSeries(pd.Series):
    def __init__(self, corner_data, real_shape):
        super().__init__(corner_data)
        self._real_shape = real_shape

    def __len__(self):
        # As we only fetch corner data to repr,
        # the length would be wrong and we have no way to control,
        # thus we just overwrite the length to show the real one
        return self._real_shape[0]


def filter_dtypes_by_index(dtypes, index):
    try:
        new_dtypes = dtypes.loc[index].dropna()
    except KeyError:
        dtypes_idx = (
            dtypes.index.to_frame()
            .merge(index.to_frame())
            .set_index(list(range(dtypes.index.nlevels)))
            .index
        )
        new_dtypes = dtypes.loc[dtypes_idx]
        new_dtypes.index.names = dtypes.index.names
    return new_dtypes


@contextmanager
def create_sa_connection(con, **kwargs):
    import sqlalchemy as sa
    from sqlalchemy.engine import Connection, Engine

    # process con
    engine = None
    if isinstance(con, Connection):
        # connection create by user
        close = False
        dispose = False
    elif isinstance(con, Engine):
        con = con.connect()
        close = True
        dispose = False
    else:
        engine = sa.create_engine(con, **kwargs)
        con = engine.connect()
        close = True
        dispose = True

    try:
        yield con
    finally:
        if close:
            con.close()
        if dispose:
            engine.dispose()


def arrow_table_to_pandas_dataframe(arrow_table, use_arrow_dtype=True, **kw):
    if not use_arrow_dtype:
        # if not use arrow string, just return
        return arrow_table.to_pandas(**kw)

    from .arrays import ArrowStringArray, ArrowListArray

    table: pa.Table = arrow_table
    schema: pa.Schema = arrow_table.schema

    arrow_field_names = list()
    arrow_arrays = list()
    arrow_indexes = list()
    other_field_names = list()
    other_arrays = list()
    for i, arrow_type in enumerate(schema.types):
        if arrow_type == pa.string() or isinstance(arrow_type, pa.ListType):
            arrow_field_names.append(schema.names[i])
            arrow_indexes.append(i)
            arrow_arrays.append(table.columns[i])
        else:
            other_field_names.append(schema.names[i])
            other_arrays.append(table.columns[i])

    df: pd.DataFrame = pa.Table.from_arrays(
        other_arrays, names=other_field_names
    ).to_pandas(**kw)
    for arrow_index, arrow_name, arrow_array in zip(
        arrow_indexes, arrow_field_names, arrow_arrays
    ):
        if arrow_array.type == pa.string():
            series = pd.Series(ArrowStringArray(arrow_array))
        else:
            assert isinstance(arrow_array.type, pa.ListType)
            series = pd.Series(ArrowListArray(arrow_array))
        df.insert(arrow_index, arrow_name, series)

    return df


def contain_arrow_dtype(dtypes):
    from .arrays import ArrowStringDtype

    return any(isinstance(dtype, ArrowStringDtype) for dtype in dtypes)


def to_arrow_dtypes(dtypes, test_df=None):
    from .arrays import ArrowStringDtype

    new_dtypes = dtypes.copy()
    for i in range(len(dtypes)):
        dtype = dtypes.iloc[i]
        if is_string_dtype(dtype):
            if test_df is not None:
                series = test_df.iloc[:, i]
                # check value
                non_na_series = series[series.notna()]
                if len(non_na_series) > 0:
                    first_value = non_na_series.iloc[0]
                    if isinstance(first_value, str):
                        new_dtypes.iloc[i] = ArrowStringDtype()
                else:  # pragma: no cover
                    # empty, set arrow string dtype
                    new_dtypes.iloc[i] = ArrowStringDtype()
            else:
                # empty, set arrow string dtype
                new_dtypes.iloc[i] = ArrowStringDtype()
    return new_dtypes


def make_dtype(dtype):
    if isinstance(dtype, (np.dtype, ExtensionDtype)):
        return dtype
    return np.dtype(dtype) if dtype is not None else None


def make_dtypes(dtypes):
    if dtypes is None:
        return None
    if not isinstance(dtypes, pd.Series):
        dtypes = pd.Series(dtypes)
    return dtypes.apply(make_dtype)


def is_dataframe(x):
    if cudf is not None:  # pragma: no cover
        if isinstance(x, cudf.DataFrame):
            return True
    return isinstance(x, pd.DataFrame)


def is_series(x):
    if cudf is not None:  # pragma: no cover
        if isinstance(x, cudf.Series):
            return True
    return isinstance(x, pd.Series)


def is_index(x):
    if cudf is not None:  # pragma: no cover
        if isinstance(x, cudf.Index):
            return True
    return isinstance(x, pd.Index)


def get_xdf(x):
    if cudf is not None:  # pragma: no cover
        if isinstance(x, (cudf.DataFrame, cudf.Series, cudf.Index)):
            return cudf
    return pd


def is_cudf(x):
    if cudf is not None:  # pragma: no cover
        if isinstance(x, (cudf.DataFrame, cudf.Series, cudf.Index)):
            return True
    return False


def auto_merge_chunks(
    ctx: Context,
    df_or_series: TileableType,
    merged_file_size: Union[int, float, str] = None,
) -> TileableType:
    from .merge import DataFrameConcat

    if df_or_series.ndim == 2 and df_or_series.chunk_shape[1] > 1:
        # skip auto merge optimization for DataFrame
        # that has more than 1 chunks on columns axis
        return df_or_series

    metas = ctx.get_chunks_meta(
        [c.key for c in df_or_series.chunks], fields=["memory_size"], error="ignore"
    )
    memory_sizes = [meta["memory_size"] if meta is not None else None for meta in metas]
    if any(size is None for size in memory_sizes):
        # has not been executed before, cannot get accurate memory size, skip auto merge
        return df_or_series

    def _concat_chunks(merge_chunks: List[ChunkType], output_index: int):
        chunk_size = sum(c.shape[0] for c in merge_chunks)
        concat_op = DataFrameConcat(output_types=df_or_series.op.output_types)
        if df_or_series.ndim == 1:
            kw = dict(
                dtype=df_or_series.dtype,
                index_value=merge_index_value(
                    {c.index: c.index_value for c in merge_chunks}
                ),
                shape=(chunk_size,),
                index=(output_index,),
                name=df_or_series.name,
            )
        else:
            kw = dict(
                dtypes=merge_chunks[0].dtypes,
                index_value=merge_index_value(
                    {c.index: c.index_value for c in merge_chunks}
                ),
                columns_value=merge_chunks[0].columns_value,
                shape=(chunk_size, merge_chunks[0].shape[1]),
                index=(output_index, 0),
            )
        return concat_op.new_chunk(merge_chunks, **kw)

    to_merge_size = (
        parse_readable_size(merged_file_size)[0]
        if merged_file_size is not None
        else options.chunk_store_limit
    )
    to_merge_chunks = []
    acc_memory_size = 0
    n_split = []
    out_chunks = []
    for chunk, chunk_memory_size in zip(df_or_series.chunks, memory_sizes):
        if (
            acc_memory_size + chunk_memory_size > to_merge_size
            and len(to_merge_chunks) > 0
        ):
            # adding current chunk would exceed the maximum,
            # concat previous chunks
            merged_chunk = _concat_chunks(to_merge_chunks, len(n_split))
            out_chunks.append(merged_chunk)
            n_split.append(merged_chunk.shape[0])
            # reset
            acc_memory_size = 0
            to_merge_chunks = []

        to_merge_chunks.append(chunk)
        acc_memory_size += chunk_memory_size
    # process the last chunk
    assert len(to_merge_chunks) >= 1
    merged_chunk = _concat_chunks(to_merge_chunks, len(n_split))
    out_chunks.append(merged_chunk)
    n_split.append(merged_chunk.shape[0])

    new_op = df_or_series.op.copy()
    params = df_or_series.params.copy()
    params["chunks"] = out_chunks
    if df_or_series.ndim == 1:
        params["nsplits"] = (tuple(n_split),)
    else:
        params["nsplits"] = (tuple(n_split), df_or_series.nsplits[1])
    return new_op.new_tileable(df_or_series.op.inputs, kws=[params])



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union

from .... import oscar as mo
from ....lib.aio import alru_cache
from ....metrics import Metrics
from ....oscar.backends.context import ProfilingContext
from ....oscar.errors import MarsError
from ....oscar.profiling import ProfilingData, MARS_ENABLE_PROFILING
from ....typing import BandType
from ....utils import dataslots, Timer
from ...subtask import Subtask, SubtaskResult, SubtaskStatus
from ...task import TaskAPI
from ..core import SubtaskScheduleSummary
from ..utils import redirect_subtask_errors

logger = logging.getLogger(__name__)


# the default times to reschedule subtask.
DEFAULT_SUBTASK_MAX_RESCHEDULES = 0


@dataslots
@dataclass
class SubtaskScheduleInfo:
    subtask: Subtask
    band_futures: Dict[BandType, asyncio.Future] = field(default_factory=dict)
    start_time: int = -1
    end_time: int = -1
    max_reschedules: int = 0
    num_reschedules: int = 0
    num_speculative_concurrent_run: int = 0

    def to_summary(self, **kwargs) -> SubtaskScheduleSummary:
        return SubtaskScheduleSummary(
            task_id=self.subtask.task_id,
            subtask_id=self.subtask.subtask_id,
            bands=list(self.band_futures.keys()),
            num_reschedules=self.num_reschedules,
            **kwargs,
        )


class SubtaskManagerActor(mo.Actor):
    _subtask_infos: Dict[str, SubtaskScheduleInfo]  # subtask id -> schedule info
    _subtask_summaries: Dict[str, SubtaskScheduleSummary]  # subtask id -> summary

    @classmethod
    def gen_uid(cls, session_id: str):
        return f"{session_id}_subtask_manager"

    def __init__(
        self,
        session_id: str,
        subtask_max_reschedules: int = DEFAULT_SUBTASK_MAX_RESCHEDULES,
        subtask_cancel_timeout: int = 5,
        speculation_config: Dict[str, object] = None,
    ):
        self._session_id = session_id
        self._subtask_infos = dict()
        self._subtask_summaries = dict()
        self._subtask_max_reschedules = subtask_max_reschedules
        self._subtask_cancel_timeout = subtask_cancel_timeout
        self._speculation_config = speculation_config or {}
        self._queueing_ref = None
        self._global_resource_ref = None
        self._submitted_subtask_count = Metrics.counter(
            "mars.scheduling.submitted_subtask_count",
            "The count of submitted subtasks to all bands.",
            ("session_id", "task_id", "stage_id"),
        )
        self._finished_subtask_count = Metrics.counter(
            "mars.scheduling.finished_subtask_count",
            "The count of finished subtasks of all bands.",
            ("session_id", "task_id", "stage_id"),
        )
        self._canceled_subtask_count = Metrics.counter(
            "mars.scheduling.canceled_subtask_count",
            "The count of canceled subtasks of all bands.",
            ("session_id", "task_id", "stage_id"),
        )
        logger.info(
            "Created SubtaskManager with subtask_max_reschedules %s, "
            "speculation_config %s",
            self._subtask_max_reschedules,
            speculation_config,
        )

    async def __post_create__(self):
        from .queueing import SubtaskQueueingActor

        self._queueing_ref = await mo.actor_ref(
            SubtaskQueueingActor.gen_uid(self._session_id), address=self.address
        )
        from ..supervisor import GlobalResourceManagerActor

        self._global_resource_ref = await mo.actor_ref(
            GlobalResourceManagerActor.default_uid(), address=self.address
        )
        from .speculation import SpeculativeScheduler

        self._speculation_execution_scheduler = SpeculativeScheduler(
            self._queueing_ref, self._global_resource_ref, self._speculation_config
        )
        await self._speculation_execution_scheduler.start()

    async def __pre_destroy__(self):
        await self._speculation_execution_scheduler.stop()

    @alru_cache
    async def _get_task_api(self):
        return await TaskAPI.create(self._session_id, self.address)

    async def add_subtasks(self, subtasks: List[Subtask], priorities: List[Tuple]):
        async with redirect_subtask_errors(self, subtasks):
            for subtask in subtasks:
                # the extra_config may be None. the extra config overwrites the default value.
                subtask_max_reschedules = (
                    subtask.extra_config.get("subtask_max_reschedules")
                    if subtask.extra_config
                    else None
                )
                if subtask_max_reschedules is None:
                    subtask_max_reschedules = self._subtask_max_reschedules
                if subtask.subtask_id in self._subtask_infos:  # pragma: no cover
                    raise KeyError(f"Subtask {subtask.subtask_id} already added.")
                self._subtask_infos[subtask.subtask_id] = SubtaskScheduleInfo(
                    subtask, max_reschedules=subtask_max_reschedules
                )

            virtual_subtasks = [subtask for subtask in subtasks if subtask.virtual]
            for subtask in virtual_subtasks:
                task_api = await self._get_task_api()
                await task_api.set_subtask_result(
                    SubtaskResult(
                        subtask_id=subtask.subtask_id,
                        session_id=subtask.session_id,
                        task_id=subtask.task_id,
                        stage_id=subtask.stage_id,
                        progress=1.0,
                        status=SubtaskStatus.succeeded,
                    )
                )
            await self._queueing_ref.add_subtasks(
                [subtask for subtask in subtasks if not subtask.virtual], priorities
            )
            await self._queueing_ref.submit_subtasks.tell()

    @alru_cache(maxsize=10000)
    async def _get_execution_ref(self, band: BandType):
        from ..worker.execution import SubtaskExecutionActor

        return await mo.actor_ref(SubtaskExecutionActor.default_uid(), address=band[0])

    async def finish_subtasks(
        self,
        subtask_ids: List[str],
        bands: List[BandType] = None,
        schedule_next: bool = True,
    ):
        logger.debug("Finished subtasks %s.", subtask_ids)
        band_tasks = defaultdict(lambda: 0)
        bands = bands or [None] * len(subtask_ids)
        for subtask_id, subtask_band in zip(subtask_ids, bands):
            subtask_info = self._subtask_infos.get(subtask_id, None)
            if subtask_info is not None:
                self._finished_subtask_count.record(
                    1,
                    {
                        "session_id": self._session_id,
                        "task_id": subtask_info.subtask.task_id,
                        "stage_id": subtask_info.subtask.stage_id,
                    },
                )
                self._subtask_summaries[subtask_id] = subtask_info.to_summary(
                    is_finished=True
                )
                subtask_info.end_time = time.time()
                self._speculation_execution_scheduler.finish_subtask(subtask_info)
                #  Cancel subtask on other bands.
                aio_task = subtask_info.band_futures.pop(subtask_band, None)
                if aio_task:
                    await aio_task
                    if schedule_next:
                        band_tasks[subtask_band] += 1
                if subtask_info.band_futures:
                    # Cancel subtask here won't change subtask status.
                    # See more in `TaskProcessorActor.set_subtask_result`
                    logger.info(
                        "Try to cancel subtask %s on bands %s.",
                        subtask_id,
                        set(subtask_info.band_futures.keys()),
                    )
                    # Cancel subtask can be async and may need to kill slot which need more time.
                    # Can't use `tell` here because next line remove subtask info which is needed by
                    # `cancel_subtasks`.
                    yield self.ref().cancel_subtasks([subtask_id])
                # cancel subtask first then pop subtask info.
                self._subtask_infos.pop(subtask_id, None)
                if schedule_next:
                    for band in subtask_info.band_futures.keys():
                        band_tasks[band] += 1
        await self._queueing_ref.remove_queued_subtasks(subtask_ids)
        if band_tasks:
            tasks = []
            for band, subtask_count in band_tasks.items():
                task = asyncio.ensure_future(
                    self._queueing_ref.submit_subtasks.tell(band, subtask_count)
                )
                tasks.append(task)
            await asyncio.wait(tasks)

    def _get_subtasks_by_ids(self, subtask_ids: List[str]) -> List[Optional[Subtask]]:
        subtasks = []
        for stid in subtask_ids:
            try:
                subtasks.append(self._subtask_infos[stid].subtask)
            except KeyError:
                subtasks.append(None)
        return subtasks

    async def submit_subtask_to_band(self, subtask_id: str, band: BandType):
        if subtask_id not in self._subtask_infos:  # pragma: no cover
            logger.info(
                "Subtask %s is not in added subtasks set, it may be finished or canceled, skip it.",
                subtask_id,
            )
            return
        async with redirect_subtask_errors(
            self, self._get_subtasks_by_ids([subtask_id])
        ):
            try:
                subtask_info = self._subtask_infos[subtask_id]
                execution_ref = await self._get_execution_ref(band)
                extra_config = subtask_info.subtask.extra_config
                enable_profiling = MARS_ENABLE_PROFILING or (
                    extra_config and extra_config.get("enable_profiling")
                )
                profiling_context = (
                    ProfilingContext(subtask_info.subtask.task_id)
                    if enable_profiling
                    else None
                )
                self._submitted_subtask_count.record(
                    1,
                    {
                        "session_id": self._session_id,
                        "task_id": subtask_info.subtask.task_id,
                        "stage_id": subtask_info.subtask.stage_id,
                    },
                )
                logger.debug("Start run subtask %s in band %s.", subtask_id, band)
                with Timer() as timer:
                    task = asyncio.create_task(
                        execution_ref.run_subtask.options(
                            profiling_context=profiling_context
                        ).send(subtask_info.subtask, band[1], self.address)
                    )
                    subtask_info.band_futures[band] = task
                    subtask_info.start_time = time.time()
                    self._speculation_execution_scheduler.add_subtask(subtask_info)
                    result = yield task
                ProfilingData.collect_subtask(
                    subtask_info.subtask, band, timer.duration
                )
                task_api = await self._get_task_api()
                logger.debug("Finished subtask %s with result %s.", subtask_id, result)
                await task_api.set_subtask_result(result)
            except (OSError, MarsError) as ex:
                # TODO: We should handle ServerClosed Error.
                if (
                    subtask_info.subtask.retryable
                    and subtask_info.num_reschedules < subtask_info.max_reschedules
                ):
                    logger.error(
                        "Reschedule subtask %s due to %s",
                        subtask_info.subtask.subtask_id,
                        ex,
                    )
                    subtask_info.num_reschedules += 1
                    await self._queueing_ref.add_subtasks(
                        [subtask_info.subtask],
                        [subtask_info.subtask.priority or tuple()],
                        exclude_bands=set(subtask_info.band_futures.keys()),
                    )
                else:
                    raise ex
            except asyncio.CancelledError:
                raise
            except Exception as ex:
                if (
                    subtask_info.subtask.retryable
                    and subtask_info.num_reschedules < subtask_info.max_reschedules
                ):
                    logger.error(
                        "Failed to reschedule subtask %s, "
                        "num_reschedules: %s, max_reschedules: %s, unhandled exception: %s",
                        subtask_info.subtask.subtask_id,
                        subtask_info.num_reschedules,
                        subtask_info.max_reschedules,
                        ex,
                    )
                raise ex
            finally:
                # make sure slot is released before marking tasks as finished
                await self._global_resource_ref.release_subtask_resource(
                    band,
                    subtask_info.subtask.session_id,
                    subtask_info.subtask.subtask_id,
                )
                logger.debug(
                    "Slot released for band %s after subtask %s",
                    band,
                    subtask_info.subtask.subtask_id,
                )
                # We should call submit_subtasks after the resource is released.
                # If submit_subtasks runs before release_subtask_resource
                # then the rescheduled subtask may not be submitted due to
                # no available resource. The mars will hangs.
                if subtask_info.num_reschedules > 0:
                    await self._queueing_ref.submit_subtasks.tell()

    async def cancel_subtasks(
        self, subtask_ids: List[str], kill_timeout: Union[float, int] = None
    ):
        kill_timeout = kill_timeout or self._subtask_cancel_timeout
        logger.info(
            "Start to cancel subtasks %s, kill timeout is %s.",
            subtask_ids,
            kill_timeout,
        )
        queued_subtask_ids = []
        single_cancel_tasks = []

        task_api = await self._get_task_api()

        async def cancel_single_task(subtask, raw_tasks, cancel_tasks):
            if cancel_tasks:
                await asyncio.wait(cancel_tasks)
            if raw_tasks:
                dones, _ = await asyncio.wait(raw_tasks)
            else:
                dones = []
            if not dones or all(fut.cancelled() for fut in dones):
                await task_api.set_subtask_result(
                    SubtaskResult(
                        subtask_id=subtask.subtask_id,
                        session_id=subtask.session_id,
                        task_id=subtask.task_id,
                        stage_id=subtask.stage_id,
                        status=SubtaskStatus.cancelled,
                    )
                )

        for subtask_id in subtask_ids:
            if subtask_id not in self._subtask_infos:
                # subtask may already finished or not submitted at all
                logger.info(
                    "Skip cancel subtask %s, it may already finished or not submitted at all",
                    subtask_id,
                )
                continue

            subtask_info = self._subtask_infos[subtask_id]
            raw_tasks_to_cancel = list(subtask_info.band_futures.values())

            if not raw_tasks_to_cancel:
                queued_subtask_ids.append(subtask_id)
                single_cancel_tasks.append(
                    asyncio.create_task(
                        cancel_single_task(subtask_info.subtask, [], [])
                    )
                )
            else:
                cancel_tasks = []
                for band in subtask_info.band_futures.keys():
                    execution_ref = await self._get_execution_ref(band)
                    cancel_tasks.append(
                        asyncio.create_task(
                            execution_ref.cancel_subtask(
                                subtask_id, kill_timeout=kill_timeout
                            )
                        )
                    )
                single_cancel_tasks.append(
                    asyncio.create_task(
                        cancel_single_task(
                            subtask_info.subtask, raw_tasks_to_cancel, cancel_tasks
                        )
                    )
                )
        if queued_subtask_ids:
            # Don't use `finish_subtasks` because it may remove queued
            await self._queueing_ref.remove_queued_subtasks(queued_subtask_ids)
        if single_cancel_tasks:
            yield asyncio.wait(single_cancel_tasks)

        for subtask_id in subtask_ids:
            subtask_info = self._subtask_infos.pop(subtask_id, None)
            if subtask_info is not None:
                self._subtask_summaries[subtask_id] = subtask_info.to_summary(
                    is_finished=True, is_cancelled=True
                )
                self._canceled_subtask_count.record(
                    1,
                    {
                        "session_id": self._session_id,
                        "task_id": subtask_info.subtask.task_id,
                        "stage_id": subtask_info.subtask.stage_id,
                    },
                )
        await self._queueing_ref.submit_subtasks.tell()
        logger.info("Subtasks %s canceled.", subtask_ids)

    def get_schedule_summaries(self, task_id: Optional[str] = None):
        if task_id is not None:
            summaries = {
                subtask_id: summary
                for subtask_id, summary in self._subtask_summaries.items()
                if summary.task_id == task_id
            }
        else:
            summaries = dict(self._subtask_summaries)
        for info in self._subtask_infos.values():
            if task_id is None or info.subtask.task_id == task_id:
                summaries[info.subtask.subtask_id] = info.to_summary()
        return list(summaries.values())



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
import operator
import weakref
from typing import Dict, List, Type, Tuple

import cloudpickle

from ...utils import no_default
from ..core import Serializer, Placeholder, buffered
from .field import Field
from .field_type import (
    PrimitiveFieldType,
    ListType,
    TupleType,
    DictType,
    DtypeType,
    DatetimeType,
    TimedeltaType,
    TZInfoType,
)


_primitive_field_types = (
    PrimitiveFieldType,
    DtypeType,
    DatetimeType,
    TimedeltaType,
    TZInfoType,
)

_is_ci = (os.environ.get("CI") or "0").lower() in ("1", "true")


def _is_field_primitive_compound(field: Field):
    if field.on_serialize is not None or field.on_deserialize is not None:
        return False

    def check_type(field_type):
        if isinstance(field_type, _primitive_field_types):
            return True
        if isinstance(field_type, (ListType, TupleType)):
            if all(
                check_type(element_type) or element_type is Ellipsis
                for element_type in field_type._field_types
            ):
                return True
        if isinstance(field_type, DictType):
            if all(
                isinstance(element_type, _primitive_field_types)
                or element_type is Ellipsis
                for element_type in (field_type.key_type, field_type.value_type)
            ):
                return True
        return False

    return check_type(field.field_type)


class SerializableMeta(type):
    def __new__(mcs, name: str, bases: Tuple[Type], properties: Dict):
        # All the fields including base fields.
        all_fields = dict()

        for base in bases:
            if hasattr(base, "_FIELDS"):
                all_fields.update(base._FIELDS)

        properties_without_fields = {}
        properties_field_slot_names = []
        for k, v in properties.items():
            if not isinstance(v, Field):
                properties_without_fields[k] = v
                continue

            field = all_fields.get(k)
            if field is None:
                properties_field_slot_names.append(k)
            else:
                v.name = field.name
                v.get = field.get
                v.set = field.set
                v.__delete__ = field.__delete__
            all_fields[k] = v

        # Make field order deterministic to serialize it as list instead of dict.
        all_fields = dict(sorted(all_fields.items(), key=operator.itemgetter(0)))
        pickle_fields = []
        non_pickle_fields = []
        for v in all_fields.values():
            if _is_field_primitive_compound(v):
                pickle_fields.append(v)
            else:
                non_pickle_fields.append(v)

        slots = set(properties.pop("__slots__", set()))
        slots.update(properties_field_slot_names)

        properties = properties_without_fields
        properties["_FIELDS"] = all_fields
        properties["_PRIMITIVE_FIELDS"] = pickle_fields
        properties["_NON_PRIMITIVE_FIELDS"] = non_pickle_fields
        properties["__slots__"] = tuple(slots)

        clz = type.__new__(mcs, name, bases, properties)
        # Bind slot member_descriptor with field.
        for name in properties_field_slot_names:
            member_descriptor = getattr(clz, name)
            field = all_fields[name]
            field.name = member_descriptor.__name__
            field.get = member_descriptor.__get__
            field.set = member_descriptor.__set__
            field.__delete__ = member_descriptor.__delete__
            setattr(clz, name, field)

        return clz


class Serializable(metaclass=SerializableMeta):
    __slots__ = ("__weakref__",)

    _cache_primitive_serial = False

    _FIELDS: Dict[str, Field]
    _PRIMITIVE_FIELDS: List[str]
    _NON_PRIMITIVE_FIELDS: List[str]

    def __init__(self, *args, **kwargs):
        fields = self._FIELDS
        if args:  # pragma: no cover
            values = dict(zip(fields, args))
            values.update(kwargs)
        else:
            values = kwargs
        for k, v in values.items():
            fields[k].set(self, v)

    def __repr__(self):
        values = ", ".join(
            [
                "{}={!r}".format(slot, getattr(self, slot, None))
                for slot in self.__slots__
            ]
        )
        return "{}({})".format(self.__class__.__name__, values)

    def copy(self) -> "Serializable":
        copied = type(self)()
        copied_fields = copied._FIELDS
        for k, field in self._FIELDS.items():
            try:
                # Slightly faster than getattr.
                value = field.get(self, k)
                copied_fields[k].set(copied, value)
            except AttributeError:
                continue
        return copied


_primitive_serial_cache = weakref.WeakKeyDictionary()


class SerializableSerializer(Serializer):
    """
    Leverage DictSerializer to perform serde.
    """

    @classmethod
    def _get_field_values(cls, obj: Serializable, fields):
        values = []
        for field in fields:
            try:
                value = field.get(obj)
                if field.on_serialize:
                    value = field.on_serialize(value)
            except AttributeError:
                # Most field values are not None, serialize by list is more efficient than dict.
                value = no_default
            values.append(value)
        return values

    @buffered
    def serial(self, obj: Serializable, context: Dict):
        if obj._cache_primitive_serial and obj in _primitive_serial_cache:
            primitive_vals = _primitive_serial_cache[obj]
        else:
            primitive_vals = self._get_field_values(obj, obj._PRIMITIVE_FIELDS)
            if obj._cache_primitive_serial:
                primitive_vals = cloudpickle.dumps(primitive_vals)
                _primitive_serial_cache[obj] = primitive_vals

        compound_vals = self._get_field_values(obj, obj._NON_PRIMITIVE_FIELDS)
        return (type(obj), primitive_vals), [compound_vals], False

    @staticmethod
    def _set_field_value(obj: Serializable, field: Field, value):
        if value is no_default:
            return
        if type(value) is Placeholder:
            if field.on_deserialize:
                value.callbacks.append(
                    lambda v: field.set(obj, field.on_deserialize(v))
                )
            else:
                value.callbacks.append(lambda v: field.set(obj, v))
        else:
            if field.on_deserialize:
                field.set(obj, field.on_deserialize(value))
            else:
                field.set(obj, value)

    def deserial(self, serialized: Tuple, context: Dict, subs: List) -> Serializable:
        obj_class, primitives = serialized

        if type(primitives) is not list:
            primitives = cloudpickle.loads(primitives)

        obj = obj_class()

        if primitives:
            for field, value in zip(obj_class._PRIMITIVE_FIELDS, primitives):
                self._set_field_value(obj, field, value)

        if obj_class._NON_PRIMITIVE_FIELDS:
            for field, value in zip(obj_class._NON_PRIMITIVE_FIELDS, subs[0]):
                self._set_field_value(obj, field, value)

        return obj


SerializableSerializer.register(Serializable)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from collections.abc import Iterable

import numpy as np

from ... import opcodes as OperandDef
from ...serialization.serializables import TupleField
from ...config import options
from ..utils import decide_chunk_sizes, gen_random_seeds
from .core import TensorRandomOperandMixin, TensorDistribution


class TensorDirichlet(TensorDistribution, TensorRandomOperandMixin):
    _op_type_ = OperandDef.RAND_DIRICHLET

    _fields_ = "alpha", "size"
    alpha = TupleField("alpha", default=None)
    _func_name = "dirichlet"

    def _calc_shape(self, shapes):
        shape = super()._calc_shape(shapes)
        return shape + (len(self.alpha),)

    def __call__(self, chunk_size=None):
        return self.new_tensor(None, None, raw_chunk_size=chunk_size)

    @classmethod
    def tile(cls, op):
        tensor = op.outputs[0]
        chunk_size = tensor.extra_params.raw_chunk_size or options.chunk_size
        nsplits = decide_chunk_sizes(
            tensor.shape[:-1], chunk_size, tensor.dtype.itemsize
        )
        nsplits += ((len(op.alpha),),)

        idxes = list(itertools.product(*[range(len(s)) for s in nsplits]))
        seeds = gen_random_seeds(len(idxes), np.random.RandomState(op.seed))

        out_chunks = []
        for seed, idx, shape in zip(seeds, idxes, itertools.product(*nsplits)):
            inputs = [inp.cix[idx] for inp in op.inputs]
            size = shape[:-1]

            chunk_op = op.copy().reset_key()
            chunk_op._state = None
            chunk_op.seed = seed
            chunk_op.size = size
            out_chunk = chunk_op.new_chunk(inputs, shape=shape, index=idx)
            out_chunks.append(out_chunk)

        new_op = op.copy()
        return new_op.new_tensors(
            op.inputs, tensor.shape, chunks=out_chunks, nsplits=nsplits
        )


def dirichlet(random_state, alpha, size=None, chunk_size=None, gpu=None, dtype=None):
    r"""
    Draw samples from the Dirichlet distribution.

    Draw `size` samples of dimension k from a Dirichlet distribution. A
    Dirichlet-distributed random variable can be seen as a multivariate
    generalization of a Beta distribution. Dirichlet pdf is the conjugate
    prior of a multinomial in Bayesian inference.

    Parameters
    ----------
    alpha : array
        Parameter of the distribution (k dimension for sample of
        dimension k).
    size : int or tuple of ints, optional
        Output shape.  If the given shape is, e.g., ``(m, n, k)``, then
        ``m * n * k`` samples are drawn.  Default is None, in which case a
        single value is returned.
    chunk_size : int or tuple of int or tuple of ints, optional
        Desired chunk size on each dimension
    gpu : bool, optional
        Allocate the tensor on GPU if True, False as default
    dtype : data-type, optional
      Data-type of the returned tensor.

    Returns
    -------
    samples : Tensor
        The drawn samples, of shape (size, alpha.ndim).

    Raises
    -------
    ValueError
        If any value in alpha is less than or equal to zero

    Notes
    -----
    .. math:: X \approx \prod_{i=1}^{k}{x^{\alpha_i-1}_i}

    Uses the following property for computation: for each dimension,
    draw a random sample y_i from a standard gamma generator of shape
    `alpha_i`, then
    :math:`X = \frac{1}{\sum_{i=1}^k{y_i}} (y_1, \ldots, y_n)` is
    Dirichlet distributed.

    References
    ----------
    .. [1] David McKay, "Information Theory, Inference and Learning
           Algorithms," chapter 23,
           http://www.inference.phy.cam.ac.uk/mackay/
    .. [2] Wikipedia, "Dirichlet distribution",
           http://en.wikipedia.org/wiki/Dirichlet_distribution

    Examples
    --------
    Taking an example cited in Wikipedia, this distribution can be used if
    one wanted to cut strings (each of initial length 1.0) into K pieces
    with different lengths, where each piece had, on average, a designated
    average length, but allowing some variation in the relative sizes of
    the pieces.

    >>> import mars.tensor as mt

    >>> s = mt.random.dirichlet((10, 5, 3), 20).transpose()

    >>> import matplotlib.pyplot as plt

    >>> plt.barh(range(20), s[0].execute())
    >>> plt.barh(range(20), s[1].execute(), left=s[0].execute(), color='g')
    >>> plt.barh(range(20), s[2].execute(), left=(s[0]+s[1]).execute(), color='r')
    >>> plt.title("Lengths of Strings")
    """
    if isinstance(alpha, Iterable):
        alpha = tuple(alpha)
    else:
        raise TypeError("`alpha` should be an array")
    if dtype is None:
        dtype = np.random.RandomState().dirichlet(alpha, size=(0,)).dtype
    size = random_state._handle_size(size)
    seed = gen_random_seeds(1, random_state.to_numpy())[0]
    op = TensorDirichlet(seed=seed, alpha=alpha, size=size, gpu=gpu, dtype=dtype)
    return op(chunk_size=chunk_size)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .initializer import DataFrame, Series, Index

# do imports to register operands
from .base.cut import cut
from .base.eval import mars_eval as eval  # pylint: disable=redefined-builtin
from .base.get_dummies import get_dummies
from .base.melt import melt
from .base.qcut import qcut
from .base.to_numeric import to_numeric
from .contrib.raydataset import to_ray_mldataset, to_ray_dataset
from .datasource.from_tensor import dataframe_from_tensor, series_from_tensor
from .datasource.from_index import series_from_index
from .datasource.from_records import from_records
from .datasource.from_vineyard import from_vineyard
from .datasource.read_csv import read_csv
from .datasource.read_sql import read_sql, read_sql_table, read_sql_query
from .datasource.read_parquet import read_parquet
from .datasource.read_raydataset import (
    read_raydataset,
    read_ray_dataset,
    read_ray_mldataset,
)
from .datasource.date_range import date_range
from .fetch import DataFrameFetch, DataFrameFetchShuffle
from .merge import concat, merge
from .missing.checkna import isna, isnull, notna, notnull
from .reduction import CustomReduction, unique
from .tseries.to_datetime import to_datetime

from . import arithmetic
from . import base
from . import indexing
from . import merge as merge_
from . import missing
from . import reduction
from . import statistics
from . import sort
from . import groupby
from . import ufunc
from . import datastore
from . import window
from . import plotting

del (
    reduction,
    statistics,
    arithmetic,
    indexing,
    merge_,
    base,
    groupby,
    missing,
    ufunc,
    datastore,
    sort,
    window,
    plotting,
)
del DataFrameFetch, DataFrameFetchShuffle

# noinspection PyUnresolvedReferences
from ..core import ExecutableTuple
from .arrays import ArrowStringDtype, ArrowStringArray, ArrowListDtype, ArrowListArray

# noinspection PyUnresolvedReferences
from pandas import Timedelta, Timestamp, offsets, NaT, Interval, DateOffset

try:
    from pandas import NA, NamedAgg
except ImportError:  # pragma: no cover
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from io import UnsupportedOperation
import logging
import sys
from typing import Dict, List, Optional, Tuple

from ..lib import sparse
from ..resource import virtual_memory
from ..utils import implements, lazy_import, calc_size_by_str
from .base import StorageBackend, StorageLevel, ObjectInfo, register_storage_backend
from .core import BufferWrappedFileObject, StorageFileObject

vineyard = lazy_import("vineyard")
vy_data_pickle = lazy_import("vineyard.data.pickle", rename="vy_data_pickle")
vy_data_utils = lazy_import("vineyard.data.utils", rename="vy_data_utils")
pyarrow = lazy_import("pyarrow")

if sys.platform.startswith("win"):
    vineyard = vy_data_pickle = vy_data_utils = None

logger = logging.getLogger(__name__)

# Setup support for mars datatypes on vineyard


def mars_sparse_matrix_builder(client, value, builder, **kw):
    meta = vineyard.ObjectMeta()
    meta["typename"] = "vineyard::SparseMatrix<%s>" % value.dtype.name
    meta["shape_"] = vy_data_utils.to_json(value.shape)
    meta.add_member("spmatrix", builder.run(client, value.spmatrix, **kw))
    return client.create_metadata(meta)


def mars_sparse_matrix_resolver(obj, resolver) -> sparse.SparseNDArray:
    meta = obj.meta
    shape = vy_data_utils.from_json(meta["shape_"])
    spmatrix = resolver.run(obj.member("spmatrix"))
    return sparse.matrix.SparseMatrix(spmatrix, shape=shape)


def _register_vineyard_matrices():
    vineyard.core.default_builder_context.register(
        sparse.matrix.SparseMatrix, mars_sparse_matrix_builder
    )
    vineyard.core.default_resolver_context.register(
        "vineyard::SparseMatrix", mars_sparse_matrix_resolver
    )


class VineyardFileObject(BufferWrappedFileObject):
    def __init__(
        self, vineyard_client, object_id, mode: str, size: Optional[int] = None
    ):
        self._client = vineyard_client
        self._file = None

        self._reader = None
        self._writer = None

        if size is None:
            size = -1  # unknown estimated size.

        super().__init__(object_id, mode, size=size)

    def _read_init(self):
        self._reader = vy_data_pickle.PickledReader(self._client.get(self._object_id))
        self._size = self._reader.store_size

    def _write_init(self):
        self._writer = vy_data_pickle.PickledWriter(self._size)

    @property
    def buffer(self):
        raise UnsupportedOperation(
            "VineyardFileObject doesn't support the direct 'buffer' property"
        )

    def read(self, size=-1):
        if not self._initialized:
            self._read_init()
            self._initialized = True
        return self._reader.read(size)

    def write(self, content: bytes):
        if not self._initialized:
            self._write_init()
            self._initialized = True
        return self._writer.write(content)

    def _read_close(self):
        self._reader = None

    def _write_close(self):
        self._writer.close()
        self._object_id = self._client.put(self._writer.value)
        self._writer = None


@register_storage_backend
class VineyardStorage(StorageBackend):
    name = "vineyard"

    def __init__(self, vineyard_size: int, vineyard_socket: str = None):
        _register_vineyard_matrices()

        self._size = vineyard_size
        self._vineyard_socket = vineyard_socket
        self._client = vineyard.connect(vineyard_socket)

    @classmethod
    @implements(StorageBackend.setup)
    async def setup(cls, **kwargs) -> Tuple[Dict, Dict]:
        loop = asyncio.get_running_loop()
        etcd_endpoints = kwargs.pop("etcd_endpoints", "127.0.0.1:2379")
        etcd_prefix = kwargs.pop("etcd_prefix", "vineyard")
        vineyard_size = kwargs.pop("vineyard_size", "1Gi")
        vineyard_socket = kwargs.pop("vineyard_socket", None)
        vineyardd_path = kwargs.pop("vineyardd_path", None)

        if kwargs:
            raise TypeError(
                f'VineyardStorage got unexpected config: {",".join(kwargs)}'
            )

        vineyard_size = calc_size_by_str(vineyard_size, virtual_memory().total)
        if vineyard_socket is not None:  # pragma: no cover
            vineyard_store = None
        else:
            vineyard_store = vineyard.deploy.local.start_vineyardd(
                etcd_endpoints,
                etcd_prefix,
                vineyardd_path,
                vineyard_size,
                vineyard_socket,
                rpc=False,
            )
            vineyard_socket = (
                await loop.run_in_executor(None, vineyard_store.__enter__)
            )[1]
        init_params = dict(vineyard_size=vineyard_size, vineyard_socket=vineyard_socket)
        teardown_params = dict(vineyard_store=vineyard_store)
        return init_params, teardown_params

    @staticmethod
    @implements(StorageBackend.teardown)
    async def teardown(**kwargs):
        vineyard_store = kwargs.get("vineyard_store")
        if vineyard_store is not None:
            vineyard_store.__exit__(None, None, None)

    @property
    @implements(StorageBackend.level)
    def level(self) -> StorageLevel:
        return StorageLevel.MEMORY

    @property
    @implements(StorageBackend.size)
    def size(self) -> Optional[int]:
        return self._size

    @property
    @implements(StorageBackend.backend_info)
    def backend_info(self):
        return {
            "name": self.name,
            "socket": self._vineyard_socket,
            "instance_id": self._client.instance_id,
        }

    @implements(StorageBackend.get)
    async def get(self, object_id, **kwargs) -> object:
        if kwargs:  # pragma: no cover
            raise NotImplementedError(f'Got unsupported args: {",".join(kwargs)}')

        return self._client.get(object_id)

    @implements(StorageBackend.put)
    async def put(self, obj, importance: int = 0) -> ObjectInfo:
        object_id = self._client.put(obj)
        size = self._client.get_meta(object_id).nbytes
        return ObjectInfo(size=size, object_id=object_id)

    @implements(StorageBackend.delete)
    async def delete(self, object_id):
        self._client.delete([object_id], deep=True)

    @implements(StorageBackend.object_info)
    async def object_info(self, object_id) -> ObjectInfo:
        size = self._client.get_meta(object_id).nbytes
        return ObjectInfo(size=size, object_id=object_id)

    @implements(StorageBackend.open_writer)
    async def open_writer(self, size=None) -> StorageFileObject:
        vineyard_writer = VineyardFileObject(self._client, None, size=size, mode="w")
        return StorageFileObject(vineyard_writer, object_id=None)

    @implements(StorageBackend.open_reader)
    async def open_reader(self, object_id) -> StorageFileObject:
        vineyard_reader = VineyardFileObject(self._client, object_id, mode="r")
        return StorageFileObject(vineyard_reader, object_id=object_id)

    @implements(StorageBackend.list)
    async def list(self) -> List:
        # FIXME: vineyard's list_objects not equal to plasma
        raise NotImplementedError



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import logging
from typing import Dict, List
from .....resource import Resource
from ..api import ExecutionConfig, register_config_cls
from ..utils import get_band_resources_from_config


logger = logging.getLogger(__name__)

IN_RAY_CI = os.environ.get("MARS_CI_BACKEND", "mars") == "ray"
# The default interval seconds to update progress and collect garbage.
DEFAULT_SUBTASK_MONITOR_INTERVAL = 0 if IN_RAY_CI else 1


@register_config_cls
class RayExecutionConfig(ExecutionConfig):
    name = "ray"

    def __init__(self, execution_config: Dict):
        super().__init__(execution_config)
        self._ray_execution_config = execution_config[self.backend]

    def get_band_resources(self):
        """
        Get the band resources from config for generating ray virtual
        resources.
        """
        return get_band_resources_from_config(self._ray_execution_config)

    def get_deploy_band_resources(self) -> List[Dict[str, Resource]]:
        return []

    def get_subtask_max_retries(self):
        return self._ray_execution_config["subtask_max_retries"]

    def get_n_cpu(self):
        return self._ray_execution_config["n_cpu"]

    def get_n_worker(self):
        return self._ray_execution_config["n_worker"]

    def get_subtask_cancel_timeout(self):
        return self._ray_execution_config["subtask_cancel_timeout"]

    def create_task_state_actor_as_needed(self):
        # Whether create RayTaskState actor as needed.
        #   - True (default):
        #     Create RayTaskState actor only when create_remote_object is called.
        #   - False:
        #     Create RayTaskState actor in advance when the RayTaskExecutor is created.
        return self._ray_execution_config.get("create_task_state_actor_as_needed", True)

    def get_subtask_monitor_interval(self):
        """
        The interval seconds for the monitor task to update progress and
        collect garbage.
        """
        return self._ray_execution_config.get(
            "subtask_monitor_interval", DEFAULT_SUBTASK_MONITOR_INTERVAL
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import tempfile
from typing import Any, Dict, List

from .... import oscar as mo
from ....lib.aio import AioFileObject


class CustomLogActor(mo.Actor):
    def __init__(self, custom_log_dir: str):
        self._custom_log_dir = custom_log_dir

    @staticmethod
    def _get_custom_log_dir(custom_log_dir: str, session_id: str):
        if custom_log_dir == "auto":
            return tempfile.mkdtemp(prefix=f"marslog-{session_id}")
        elif custom_log_dir is None:
            return
        else:
            return os.path.join(custom_log_dir, session_id)

    def new_custom_log_dir(self, session_id: str):
        custom_log_dir = self._get_custom_log_dir(self._custom_log_dir, session_id)
        if custom_log_dir:
            os.makedirs(custom_log_dir, exist_ok=True)
            return custom_log_dir

    @classmethod
    def clear_custom_log_dirs(cls, paths: List[str]):
        _ = [shutil.rmtree(path, ignore_errors=True) for path in paths]

    @classmethod
    async def fetch_logs(
        cls, log_paths: List[str], offsets: List[int], sizes: List[int]
    ) -> List[Dict[str, Any]]:
        result = []
        for i, log_path in enumerate(log_paths):
            log_result = dict()

            offset = offsets[i]
            size = sizes[i]

            async with AioFileObject(open(log_path)) as f:
                if offset < 0:
                    # process negative offset
                    offset = max(os.path.getsize(log_path) + offset, 0)

                if offset:
                    await f.seek(offset)

                log_result["log"] = await f.read(size)
                log_result["offset"] = await f.tell()

            result.append(log_result)

        return result



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

import numpy as np

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE, recursive_tile
from ...core.context import get_context
from ...serialization.serializables import KeyField, AnyField, StringField, BoolField
from ...utils import has_unknown_shape
from ..datasource import tensor as astensor
from ..base import moveaxis, where
from ..indexing import take
from ..arithmetic import isnan, add
from ..reduction import any as tensor_any
from ..operands import TensorOperand, TensorOperandMixin
from ..core import TENSOR_TYPE, TENSOR_CHUNK_TYPE, TensorOrder
from ..utils import check_out_param
from ..array_utils import as_same_device, device
from .core import _ureduce


def _quantile_is_valid(q):
    # avoid expensive reductions, relevant for arrays with < O(1000) elements
    if q.ndim == 1 and q.size < 10:
        for i in range(q.size):
            if q[i] < 0.0 or q[i] > 1.0:
                return False
    else:
        # faster than any()
        if np.count_nonzero(q < 0.0) or np.count_nonzero(q > 1.0):
            return False
    return True


def _quantile_ureduce_func(
    a,
    q,
    axis=None,
    out=None,
    overwrite_input=False,
    interpolation="linear",
    keepdims=False,
):
    a = astensor(a)
    out = astensor(out) if out is not None else None

    if q.ndim == 0:
        # Do not allow 0-d arrays because following code fails for scalar
        zerod = True
        q = q[None]
    else:
        zerod = False

    # prepare a for partitioning
    if overwrite_input:
        if axis is None:
            ap = a.ravel()
        else:
            ap = a
    else:
        if axis is None:
            ap = a.flatten()
        else:
            ap = a.copy()

    if axis is None:
        axis = 0

    Nx = ap.shape[axis]
    indices = q * (Nx - 1)

    # round fractional indices according to interpolation method
    if interpolation == "lower":
        indices = np.floor(indices).astype(np.intp)
    elif interpolation == "higher":
        indices = np.ceil(indices).astype(np.intp)
    elif interpolation == "midpoint":
        indices = 0.5 * (np.floor(indices) + np.ceil(indices))
    elif interpolation == "nearest":
        indices = np.around(indices).astype(np.intp)
    else:
        assert interpolation == "linear"
        # keep index as fraction and interpolate

    n = np.array(False, dtype=bool)  # check for nan's flag
    if indices.dtype == np.intp:  # take the points along axis
        # Check if the array contains any nan's
        if np.issubdtype(a.dtype, np.inexact):
            indices = np.concatenate((indices, [-1]))

        ap.partition(indices, axis=axis, need_align=True)
        # ensure axis with q-th is first
        ap = moveaxis(ap, axis, 0)
        axis = 0

        # Check if the array contains any nan's
        if np.issubdtype(a.dtype, np.inexact):
            indices = indices[:-1]
            n = isnan(ap[-1:, ...])

        if zerod:
            indices = indices[0]
        r = take(ap, indices, axis=axis, out=out)

    else:  # weight the points above and below the indices
        indices_below = np.floor(indices).astype(np.intp)
        indices_above = indices_below + 1
        indices_above[indices_above > Nx - 1] = Nx - 1

        # Check if the array contains any nan's
        if np.issubdtype(a.dtype, np.inexact):
            indices_above = np.concatenate((indices_above, [-1]))

        weights_above = indices - indices_below
        weights_below = 1 - weights_above

        weights_shape = [1] * ap.ndim
        weights_shape[axis] = len(indices)
        weights_below.shape = weights_shape
        weights_above.shape = weights_shape

        ap.partition(
            np.concatenate((indices_below, indices_above)), axis=axis, need_align=True
        )

        # ensure axis with q-th is first
        ap = moveaxis(ap, axis, 0)
        weights_below = np.moveaxis(weights_below, axis, 0)
        weights_above = np.moveaxis(weights_above, axis, 0)
        axis = 0

        # Check if the array contains any nan's
        if np.issubdtype(a.dtype, np.inexact):
            indices_above = indices_above[:-1]
            n = isnan(ap[-1:, ...])

        x1 = take(ap, indices_below, axis=axis) * weights_below
        x2 = take(ap, indices_above, axis=axis) * weights_above

        # ensure axis with q-th is first
        x1 = moveaxis(x1, axis, 0)
        x2 = moveaxis(x2, axis, 0)

        if zerod:
            x1 = x1.squeeze(0)
            x2 = x2.squeeze(0)

        if out is not None:
            r = add(x1, x2, out=out)
        else:
            r = add(x1, x2)

    if isinstance(n, TENSOR_TYPE):
        if zerod:
            if ap.ndim == 1:
                r.data = where(tensor_any(n), a.dtype.type(np.nan), r).data
                if out is not None:
                    out.data = r.data
            else:
                r[:] = where(
                    tensor_any(n), where(n.squeeze(0), a.dtype.type(np.nan), r), r
                )
        else:
            if r.ndim == 1:
                r[:] = where(tensor_any(n), np.full(r.shape, a.dtype.type(np.nan)), r)
            else:
                r[:] = where(
                    tensor_any(n),
                    where(n.repeat(q.size, 0), a.dtype.type(np.nan), r),
                    r,
                )

    return r


q_error_msg = "Quantiles must be in the range [0, 1]"


class TensorQuantile(TensorOperand, TensorOperandMixin):
    __slots__ = ("q_error_msg",)
    _op_type_ = OperandDef.QUANTILE

    _a = KeyField("a")
    _q = AnyField("q")
    _axis = AnyField("axis")
    _out = KeyField("out")
    _overwrite_input = BoolField("overwrite_input")
    _interpolation = StringField("interpolation")
    _keepdims = BoolField("keepdims")

    def __init__(
        self,
        q=None,
        axis=None,
        out=None,
        overwrite_input=None,
        interpolation=None,
        keepdims=None,
        **kw,
    ):
        self.q_error_msg = kw.pop("q_error_msg", q_error_msg)
        super().__init__(
            _q=q,
            _axis=axis,
            _interpolation=interpolation,
            _out=out,
            _overwrite_input=overwrite_input,
            _keepdims=keepdims,
            **kw,
        )

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        self._a = self._inputs[0]
        if isinstance(self._q, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._q = self._inputs[1]
        if isinstance(self._out, (TENSOR_TYPE, TENSOR_CHUNK_TYPE)):
            self._out = self._inputs[-1]

    @property
    def a(self):
        return self._a

    @property
    def q(self):
        return self._q

    @property
    def axis(self):
        return self._axis

    @property
    def out(self):
        return self._out

    @property
    def overwrite_input(self):
        return self._overwrite_input

    @property
    def interpolation(self):
        return self._interpolation

    @property
    def keepdims(self):
        return self._keepdims

    def __call__(self, a, q=None, out=None):
        shape = [self._q.size] if self._q.ndim > 0 else []
        if self._axis is None:
            exclude_axes = set(range(a.ndim))
        elif isinstance(self._axis, tuple):
            exclude_axes = set(self._axis)
        else:
            exclude_axes = {self._axis}
        for ax, s in enumerate(a.shape):
            if ax not in exclude_axes:
                shape.append(s)
            elif self._keepdims:
                shape.append(1)
        inputs = [a] if q is None else [a, q]
        order = TensorOrder.C_ORDER
        if out is not None:
            inputs.append(out)
            order = out.order
            shape = out.shape
        t = self.new_tensor(inputs, shape=tuple(shape), order=order)
        if out is not None:
            check_out_param(out, t, "same_kind")
            out.data = t.data
            return out
        else:
            return t

    @classmethod
    def _tile(cls, op, q):
        r, k = _ureduce(
            op.a,
            func=_quantile_ureduce_func,
            q=q,
            axis=op.axis,
            out=op.out,
            overwrite_input=op.overwrite_input,
            interpolation=op.interpolation,
        )
        if op.keepdims:
            return r.reshape(q.shape + k)
        else:
            return r

    @classmethod
    def _tile_one_chunk(cls, op, q):
        in_tensor = op.inputs[0]
        out_tensor = op.outputs[0]
        chunk_op = op.copy().reset_key()
        chunk_op._q = q
        chunk_inputs = [in_tensor.chunks[0]]
        if op.out is not None:
            chunk_inputs.append(op.out.chunks[0])
        chunk = chunk_op.new_chunk(
            chunk_inputs,
            shape=out_tensor.shape,
            index=(0,) * out_tensor.ndim,
            order=out_tensor.order,
        )
        op = op.copy()
        return op.new_tensors(
            op.inputs,
            shape=out_tensor.shape,
            order=out_tensor.order,
            nsplits=tuple((s,) for s in out_tensor.shape),
            chunks=[chunk],
        )

    @classmethod
    def tile(cls, op):
        if isinstance(op.q, TENSOR_TYPE):
            ctx = get_context()
            # get q's data
            q_chunk_keys = [c.key for c in op.q.chunks]
            try:
                q_data = ctx.get_chunks_result(q_chunk_keys)
            except KeyError:
                # trigger execution of `q`
                yield op.q.chunks
                q_data = ctx.get_chunks_result(q_chunk_keys)
            op._q = q = np.concatenate(q_data)
            if not _quantile_is_valid(q):
                raise ValueError(op.q_error_msg)
        else:
            if has_unknown_shape(*op.inputs):
                yield
            q = np.asarray(op.q)

        if len(op.a.chunks) == 1 and (op.out is None or len(op.out.chunks) == 1):
            return cls._tile_one_chunk(op, q)
        else:
            tiled = yield from recursive_tile(cls._tile(op, q))
            return [tiled]

    @classmethod
    def execute(cls, ctx, op):
        inputs, device_id, xp = as_same_device(
            [ctx[inp.key] for inp in op.inputs], device=op.device, ret_extra=True
        )
        a = inputs[0]
        out = inputs[-1].copy() if op.out is not None else None

        with device(device_id):
            ctx[op.outputs[0].key] = xp.quantile(
                a,
                q=op.q,
                axis=op.axis,
                out=out,
                interpolation=op.interpolation,
                keepdims=op.keepdims,
            )


INTERPOLATION_TYPES = {"linear", "lower", "higher", "midpoint", "nearest"}


def _quantile_unchecked(
    a,
    q,
    axis=None,
    out=None,
    overwrite_input=False,
    interpolation="linear",
    keepdims=False,
    q_error_msg=None,
    handle_non_numeric=None,
):
    a = astensor(a)
    raw_dtype = a.dtype
    need_view_back = False
    if handle_non_numeric and not np.issubdtype(a.dtype, np.number):
        # enable handle_non_numeric is often used
        # to handle the datetime-like dtype
        a = a.astype("i8")
        need_view_back = True
    if isinstance(q, ENTITY_TYPE):
        q = astensor(q)
        # do check in tile
        q_input = q
    else:
        q_input = None

    if isinstance(axis, Iterable):
        axis = tuple(axis)

    if q.ndim > 1:
        raise ValueError("`q` should be a scalar or array of float")

    if out is not None and not isinstance(out, TENSOR_TYPE):
        raise TypeError(f"`out` should be a tensor, got {type(out)}")

    if interpolation not in INTERPOLATION_TYPES:
        raise ValueError(
            "interpolation can only be 'linear', 'lower' "
            "'higher', 'midpoint', or 'nearest'"
        )

    # infer dtype
    q_tiny = np.random.rand(2 if q.size % 2 == 0 else 1).astype(q.dtype)
    if handle_non_numeric and not np.issubdtype(a.dtype, np.number):
        dtype = a.dtype
    else:
        dtype = np.quantile(
            np.empty(1, dtype=a.dtype), q_tiny, interpolation=interpolation
        ).dtype
    op = TensorQuantile(
        q=q,
        axis=axis,
        out=out,
        overwrite_input=overwrite_input,
        interpolation=interpolation,
        keepdims=keepdims,
        handle_non_numeric=handle_non_numeric,
        q_error_msg=q_error_msg,
        dtype=dtype,
        gpu=a.op.gpu,
    )
    ret = op(a, q=q_input, out=out)
    if need_view_back:
        ret = ret.astype(raw_dtype)
    return ret


def quantile(
    a,
    q,
    axis=None,
    out=None,
    overwrite_input=False,
    interpolation="linear",
    keepdims=False,
    **kw,
):
    """
    Compute the q-th quantile of the data along the specified axis.

    Parameters
    ----------
    a : array_like
        Input tensor or object that can be converted to a tensor.
    q : array_like of float
        Quantile or sequence of quantiles to compute, which must be between
        0 and 1 inclusive.
    axis : {int, tuple of int, None}, optional
        Axis or axes along which the quantiles are computed. The
        default is to compute the quantile(s) along a flattened
        version of the tensor.
    out : Tensor, optional
        Alternative output tensor in which to place the result. It must
        have the same shape and buffer length as the expected output,
        but the type (of the output) will be cast if necessary.
    overwrite_input : bool, optional
        Just for compatibility with Numpy, would not take effect.
    interpolation : {'linear', 'lower', 'higher', 'midpoint', 'nearest'}
        This optional parameter specifies the interpolation method to
        use when the desired quantile lies between two data points
        ``i < j``:

            * linear: ``i + (j - i) * fraction``, where ``fraction``
              is the fractional part of the index surrounded by ``i``
              and ``j``.
            * lower: ``i``.
            * higher: ``j``.
            * nearest: ``i`` or ``j``, whichever is nearest.
            * midpoint: ``(i + j) / 2``.
    keepdims : bool, optional
        If this is set to True, the axes which are reduced are left in
        the result as dimensions with size one. With this option, the
        result will broadcast correctly against the original tensor `a`.

    Returns
    -------
    quantile : scalar or Tensor
        If `q` is a single quantile and `axis=None`, then the result
        is a scalar. If multiple quantiles are given, first axis of
        the result corresponds to the quantiles. The other axes are
        the axes that remain after the reduction of `a`. If the input
        contains integers or floats smaller than ``float64``, the output
        data-type is ``float64``. Otherwise, the output data-type is the
        same as that of the input. If `out` is specified, that tensor is
        returned instead.

    See Also
    --------
    mean
    percentile : equivalent to quantile, but with q in the range [0, 100].
    median : equivalent to ``quantile(..., 0.5)``
    nanquantile

    Notes
    -----
    Given a vector ``V`` of length ``N``, the q-th quantile of
    ``V`` is the value ``q`` of the way from the minimum to the
    maximum in a sorted copy of ``V``. The values and distances of
    the two nearest neighbors as well as the `interpolation` parameter
    will determine the quantile if the normalized ranking does not
    match the location of ``q`` exactly. This function is the same as
    the median if ``q=0.5``, the same as the minimum if ``q=0.0`` and the
    same as the maximum if ``q=1.0``.

    Examples
    --------
    >>> import mars.tensor as mt
    >>> a = mt.array([[10, 7, 4], [3, 2, 1]])
    >>> a.execute()
    array([[10,  7,  4],
           [ 3,  2,  1]])
    >>> mt.quantile(a, 0.5).execute()
    3.5
    >>> mt.quantile(a, 0.5, axis=0).execute()
    array([6.5, 4.5, 2.5])
    >>> mt.quantile(a, 0.5, axis=1).execute()
    array([7.,  2.])
    >>> mt.quantile(a, 0.5, axis=1, keepdims=True).execute()
    array([[7.],
           [2.]])
    >>> m = mt.quantile(a, 0.5, axis=0)
    >>> out = mt.zeros_like(m)
    >>> mt.quantile(a, 0.5, axis=0, out=out).execute()
    array([6.5, 4.5, 2.5])
    >>> m.execute()
    array([6.5, 4.5, 2.5])
    """

    handle_non_numeric = kw.pop("handle_non_numeric", None)
    if len(kw) > 0:  # pragma: no cover
        raise TypeError(
            f"quantile() got an unexpected keyword argument '{next(iter(kw))}'"
        )

    if not isinstance(q, ENTITY_TYPE):
        q = np.asanyarray(q)
        # do check instantly if q is not a tensor
        if not _quantile_is_valid(q):
            raise ValueError(q_error_msg)

    return _quantile_unchecked(
        a,
        q,
        axis=axis,
        out=out,
        overwrite_input=overwrite_input,
        interpolation=interpolation,
        keepdims=keepdims,
        handle_non_numeric=handle_non_numeric,
    )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numbers import Integral
from typing import Dict

import numpy as np
import pandas as pd
from pandas.core.dtypes.cast import find_common_type
from pandas.core.indexing import IndexingError

from ... import opcodes as OperandDef
from ...core import ENTITY_TYPE
from ...core.operand import OperandStage
from ...serialization.serializables import KeyField, ListField
from ...tensor.datasource import asarray
from ...tensor.utils import calc_sliced_size, filter_inputs
from ...utils import lazy_import, is_full_slice
from ..core import IndexValue, DATAFRAME_TYPE
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index
from .index_lib import DataFrameLocIndexesHandler


cudf = lazy_import("cudf")


def process_loc_indexes(inp, indexes):
    ndim = inp.ndim

    if not isinstance(indexes, tuple):
        indexes = (indexes,)
    if len(indexes) < ndim:
        indexes += (slice(None),) * (ndim - len(indexes))
    if len(indexes) > ndim:
        raise IndexingError("Too many indexers")

    new_indexes = []
    for ax, index in enumerate(indexes):
        if isinstance(index, (list, np.ndarray, pd.Series, ENTITY_TYPE)):
            if not isinstance(index, ENTITY_TYPE):
                index = np.asarray(index)
            else:
                index = asarray(index)
                if ax == 1:
                    # do not support tensor index on axis 1
                    # because if so, the dtypes and columns_value would be unknown
                    try:
                        index = index.fetch()
                    except (RuntimeError, ValueError):
                        raise NotImplementedError(
                            "indexer on axis columns cannot be non-executed tensor"
                        )
        new_indexes.append(index)

    return new_indexes


class DataFrameLoc:
    def __init__(self, obj):
        self._obj = obj

    def _use_iloc(self, indexes):
        # for RangeIndex from 0, use iloc instead of loc
        index_value = self._obj.index_value.value
        if len(indexes) == 2:
            if not isinstance(indexes[1], slice):
                return False
            elif indexes[1] != slice(None):
                return False
        if not isinstance(index_value, IndexValue.RangeIndex):
            return False
        if index_value.slice.start != 0 and index_value.slice.start is not None:
            return False
        if not isinstance(indexes[0], (Integral, slice)):
            return False
        if isinstance(indexes[0], Integral):
            if indexes[0] < 0:
                return False
        else:
            for v in (indexes[0].start, indexes[0].stop, indexes[0].step):
                if v is None:
                    continue
                if not isinstance(v, Integral):
                    return False
                if v < 0:
                    return False
        return True

    def __getitem__(self, indexes):
        indexes = process_loc_indexes(self._obj, indexes)

        if self._use_iloc(indexes):
            # use iloc instead
            return self._obj.iloc[tuple(indexes)]

        op = DataFrameLocGetItem(indexes=indexes)
        return op(self._obj)


class DataFrameLocGetItem(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DATAFRAME_LOC_GETITEM

    _input = KeyField("input")
    _indexes = ListField("indexes")

    def __init__(self, indexes=None, gpu=None, sparse=False, output_types=None, **kw):
        super().__init__(
            _indexes=indexes, gpu=gpu, sparse=sparse, _output_types=output_types, **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def indexes(self):
        return self._indexes

    @property
    def can_index_miss(self):
        return False

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        inputs_iter = iter(self._inputs)
        self._input = next(inputs_iter)
        indexes = []
        for index in self._indexes:
            if isinstance(index, ENTITY_TYPE):
                indexes.append(next(inputs_iter))
            else:
                indexes.append(index)
        self._indexes = list(indexes)

    @classmethod
    def _calc_slice_param(
        cls,
        input_index_value: IndexValue,
        pd_index: pd.Index,
        inp,
        index: slice,
        axis: int,
    ) -> Dict:
        param = dict()
        if is_full_slice(index):
            # full slice on this axis
            param["shape"] = inp.shape[axis]
            param["index_value"] = input_index_value
            if axis == 1:
                param["dtypes"] = inp.dtypes
        elif input_index_value.has_value():
            start, end = pd_index.slice_locs(
                index.start, index.stop, index.step, kind="loc"
            )
            slc = slice(start, end, index.step)
            size = calc_sliced_size(inp.shape[axis], slc)
            param["shape"] = size
            out_index = pd_index[slc]
            param["index_value"] = parse_index(out_index, store_data=axis == 1)
            if axis == 1:
                param["dtypes"] = inp.dtypes[slc]
        else:
            assert axis == 0
            if index.start is None and index.stop is None:
                param["shape"] = calc_sliced_size(inp.shape[axis], index)
            else:
                param["shape"] = np.nan
            param["index_value"] = parse_index(pd_index, inp, index)

        return param

    @classmethod
    def _calc_bool_index_param(
        cls, input_index_value: IndexValue, pd_index: pd.Index, inp, index, axis: int
    ) -> Dict:
        param = dict()
        if input_index_value.has_value():
            if isinstance(index, np.ndarray):
                filtered_index = pd_index[index]
                param["shape"] = len(filtered_index)
                param["index_value"] = parse_index(filtered_index, store_data=axis == 1)
                if axis == 1:
                    param["dtypes"] = inp.dtypes[index]
            else:
                # tensor, cannot be indexer on axis 1
                assert axis == 0
                param["shape"] = np.nan
                param["index_value"] = parse_index(
                    pd.Index([], dtype=pd_index.dtype), inp, index, store_data=False
                )
        else:
            assert axis == 0
            if isinstance(index, np.ndarray):
                param["shape"] = int(index.sum())
            else:
                param["shape"] = np.nan
            param["index_value"] = parse_index(pd_index, inp, index, store_data=False)

        return param

    @classmethod
    def _calc_fancy_index_param(
        cls, input_index_value: IndexValue, pd_index: pd.Index, inp, index, axis: int
    ) -> Dict:
        param = dict()
        if input_index_value.has_value():
            if isinstance(index, np.ndarray):
                if not pd_index.is_unique:
                    assert axis == 1
                    # as there's no direct method in pandas to handle fancy indexes
                    # we creates a empty
                    new_dtypes = inp.dtypes.loc[index]
                    param["shape"] = len(new_dtypes)
                    param["index_value"] = parse_index(
                        new_dtypes.index, store_data=True
                    )
                    param["dtypes"] = new_dtypes
                else:
                    for it in index:
                        if it not in pd_index:
                            axis_name = "index" if axis == 0 else "columns"
                            raise KeyError(
                                f"Label [{it}] not found in the [{axis_name}]"
                            )
                    param["shape"] = len(index)
                    param["index_value"] = parse_index(pd.Index(index), store_data=True)
                    if axis == 1:
                        param["dtypes"] = inp.dtypes[index]
            else:
                assert axis == 0
                param["shape"] = index.shape[0]
                param["index_value"] = parse_index(
                    pd.Index([], dtype=pd_index.dtype), inp, index
                )
        else:
            assert axis == 0
            param["shape"] = index.shape[0]
            param["index_value"] = parse_index(pd_index, inp, index)

        return param

    @classmethod
    def _calc_param(cls, inp, axis: int, index) -> Dict:
        input_index_value = inp.index_value if axis == 0 else inp.columns_value
        pd_index = input_index_value.to_pandas()

        if isinstance(index, slice):
            return cls._calc_slice_param(input_index_value, pd_index, inp, index, axis)
        elif hasattr(index, "dtype") and index.ndim == 1:
            if index.dtype == np.bool_:
                # bool indexing
                return cls._calc_bool_index_param(
                    input_index_value, pd_index, inp, index, axis
                )
            else:
                # fancy indexing
                return cls._calc_fancy_index_param(
                    input_index_value, pd_index, inp, index, axis
                )
        else:
            param = dict()
            if input_index_value.has_value():
                loc = pd_index.get_loc(index)
                if isinstance(loc, (slice, np.ndarray)):
                    assert axis == 1
                    new_dtypes = inp.dtypes[loc]
                    param["shape"] = len(new_dtypes)
                    param["index_value"] = parse_index(
                        new_dtypes.index, store_data=True
                    )
                    param["dtypes"] = new_dtypes
                else:
                    # append None to indicate returning Series
                    param["shape"] = None
            else:
                param["shape"] = None
            return param

    def __call__(self, inp):
        inputs = [inp] + filter_inputs(self._indexes)

        shape = []
        sizes = []
        index_value = columns_value = dtypes = None
        for ax, index in enumerate(self._indexes):
            param = self._calc_param(inp, ax, index)

            size = param.get("shape")
            sizes.append(size)
            if size is not None:
                shape.append(size)

            if ax == 0:
                index_value = param.get("index_value")
            else:
                columns_value = param.get("index_value")
                dtypes = param.get("dtypes")

        shape = tuple(shape)
        if len(shape) == 0:
            # scalar
            if isinstance(inp, DATAFRAME_TYPE):
                dtype = inp.dtypes[self._indexes[1]]
            else:
                dtype = inp.dtype
            return self.new_scalar(inputs, dtype=dtype)
        elif len(shape) == 1:
            # series
            if isinstance(inp, DATAFRAME_TYPE):
                if sizes[0] is None:
                    # label on axis 0
                    dtype = find_common_type(list(dtypes))
                    return self.new_series(
                        inputs,
                        shape=shape,
                        dtype=dtype,
                        index_value=columns_value,
                        name=self._indexes[0],
                    )
                else:
                    # label on axis 1
                    dtype = inp.dtypes[self._indexes[1]]
                    return self.new_series(
                        inputs,
                        shape=shape,
                        dtype=dtype,
                        index_value=index_value,
                        name=self._indexes[1],
                    )
            else:
                return self.new_series(
                    inputs,
                    shape=shape,
                    dtype=inp.dtype,
                    index_value=index_value,
                    name=inp.name,
                )
        else:
            # dataframe
            return self.new_dataframe(
                inputs,
                shape=shape,
                dtypes=dtypes,
                index_value=index_value,
                columns_value=columns_value,
            )

    @classmethod
    def tile(cls, op):
        handler = DataFrameLocIndexesHandler()
        return [(yield from handler.handle(op))]

    @classmethod
    def execute(cls, ctx, op):
        chunk = op.outputs[0]
        df = ctx[op.input.key]
        if len(op.inputs) > 1:
            indexes = tuple(
                ctx[index.key] if hasattr(index, "key") else index
                for index in op.indexes
            )
        else:
            indexes = tuple(op.indexes)
        xdf = pd if isinstance(df, (pd.Series, pd.DataFrame)) or cudf is None else cudf

        if op.stage != OperandStage.map:
            try:
                r = df.loc[indexes]
            except AttributeError:
                # workaround for error when calling series.loc[(index,)]
                r = df.loc[indexes[0]]
        else:
            # for map stage, and when some index is fancy index
            # ignore keys that do not exist
            new_indexes = []
            str_loc_on_datetime_index = False
            for ax, index in enumerate(indexes):
                if ax == 0:
                    if isinstance(index, np.ndarray) and index.dtype != np.bool_:
                        new_indexes.append(df.index.intersection(index))
                    elif isinstance(df.index, pd.DatetimeIndex) and isinstance(
                        index, str
                    ):
                        # special process for datetime index
                        str_loc_on_datetime_index = True
                        new_indexes.append(index)
                    else:
                        new_indexes.append(index)
                else:
                    new_indexes.append(index)

            try:
                r = df.loc[tuple(new_indexes)]
                if str_loc_on_datetime_index:
                    # convert back to DataFrame or Series
                    if r.ndim == 0:
                        index = df.index[df.index.get_loc(new_indexes[0])]
                        r = xdf.Series([r], index=[index])
                    elif r.ndim == 1:
                        rdf = xdf.DataFrame(columns=r.index)
                        rdf.loc[r.name] = r
                        r = rdf
            except KeyError:
                if str_loc_on_datetime_index:
                    new_indexes[0] = []
                    r = df.loc[tuple(new_indexes)]
                else:  # pragma: no cover
                    raise

        if isinstance(r, pd.Series) and r.dtype != chunk.dtype:
            r = r.astype(chunk.dtype)
        ctx[chunk.key] = r


def loc(a):
    return DataFrameLoc(a)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import importlib
import logging
import os

from tornado import web

from ... import oscar as mo
from ...utils import get_next_port
from ..core import AbstractService

logger = logging.getLogger(__name__)


class WebActor(mo.Actor):
    def __init__(self, config):
        super().__init__()
        self._config = config
        self._web_server = None
        self._web_app = None

        extra_mod_names = self._config.get("extra_discovery_modules") or []
        web_handlers = self._config.get("web_handlers", {})
        for mod_name in extra_mod_names:
            try:
                web_mod = importlib.import_module(mod_name)
                web_handlers.update(getattr(web_mod, "web_handlers", {}))
            except ImportError:  # pragma: no cover
                pass

    async def __post_create__(self):
        from .indexhandler import handlers as web_handlers

        static_path = os.path.join(os.path.dirname(__file__), "static")
        supervisor_addr = self.address

        host = self._config.get("host") or "0.0.0.0"
        port = self._config.get("port") or get_next_port()
        self._web_address = f"http://{host}:{port}"
        web_handlers.update(self._config.get("web_handlers", {}))

        handler_kwargs = {"supervisor_addr": supervisor_addr}
        handlers = [
            (r"[^\?\&]*/static/(.*)", web.StaticFileHandler, {"path": static_path})
        ]
        for p, h in web_handlers.items():
            handlers.append((p, h, handler_kwargs))

        retrial = 5
        while retrial:
            try:
                if port is None:
                    port = get_next_port()

                self._web_app = web.Application(handlers)
                self._web_server = self._web_app.listen(port, host)
                logger.info("Mars Web started at %s:%d", host, port)
                break
            except OSError:  # pragma: no cover
                if port is not None:
                    raise
                retrial -= 1
                if retrial == 0:
                    raise

    async def __pre_destroy__(self):
        if self._web_server is not None:
            self._web_server.stop()

    def get_web_address(self):
        web_address = self._web_address
        if os.name == "nt":
            web_address = web_address.replace("0.0.0.0", "127.0.0.1")
        return web_address


class WebSupervisorService(AbstractService):
    """
    Web service on supervisor.

    Service Configuration
    ---------------------
    {
        "web": {
            "host": "<web host>",
            "port": "<web port>",
            "web_handlers": [
                <web_handlers>,
            ],
            "extra_discovery_modules": [
                "path.to.modules",
            ]
        }
    }
    """

    async def start(self):
        await mo.create_actor(
            WebActor,
            config=self._config.get("web", {}),
            uid=WebActor.default_uid(),
            address=self._address,
        )

    async def stop(self):
        await mo.destroy_actor(
            mo.create_actor_ref(uid=WebActor.default_uid(), address=self._address)
        )



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import pandas as pd

from ... import opcodes as OperandDef
from ... import tensor as mt
from ...core import recursive_tile
from ...core.operand import OperandStage
from ...serialization.serializables import FieldTypes, KeyField, ListField, AnyField
from ...utils import lazy_import, has_unknown_shape
from ..core import SERIES_TYPE
from ..initializer import DataFrame, Series
from ..operands import DataFrameOperand, DataFrameOperandMixin
from ..utils import parse_index, build_empty_df


cudf = lazy_import("cudf")


class DataFrameDescribe(DataFrameOperand, DataFrameOperandMixin):
    _op_type_ = OperandDef.DESCRIBE

    _input = KeyField("input")
    _percentiles = ListField("percentiles", FieldTypes.float64)
    _include = AnyField("include")
    _exclude = AnyField("exclude")

    def __init__(
        self, percentiles=None, include=None, exclude=None, output_types=None, **kw
    ):
        super().__init__(
            _percentiles=percentiles,
            _include=include,
            _exclude=exclude,
            _output_types=output_types,
            **kw
        )

    @property
    def input(self):
        return self._input

    @property
    def percentiles(self):
        return self._percentiles

    @property
    def include(self):
        return self._include

    @property
    def exclude(self):
        return self._exclude

    def _set_inputs(self, inputs):
        super()._set_inputs(inputs)
        if self.stage != OperandStage.agg:
            self._input = self._inputs[0]

    def __call__(self, df_or_series):
        if isinstance(df_or_series, SERIES_TYPE):
            if not np.issubdtype(df_or_series.dtype, np.number):
                raise NotImplementedError("non-numeric type is not supported for now")
            test_series = pd.Series([], dtype=df_or_series.dtype).describe(
                percentiles=self._percentiles,
                include=self._include,
                exclude=self._exclude,
            )
            return self.new_series(
                [df_or_series],
                shape=(len(test_series),),
                dtype=test_series.dtype,
                index_value=parse_index(test_series.index, store_data=True),
            )
        else:
            test_inp_df = build_empty_df(df_or_series.dtypes)
            test_df = test_inp_df.describe(
                percentiles=self._percentiles,
                include=self._include,
                exclude=self._exclude,
            )
            if len(self.percentiles) == 0:
                # specify percentiles=False
                # Note: unlike pandas that False is illegal value for percentiles,
                # Mars DataFrame allows user to specify percentiles=False
                # to skip computation about percentiles
                test_df.drop(["50%"], axis=0, inplace=True)
            for dtype in test_df.dtypes:
                if not np.issubdtype(dtype, np.number):
                    raise NotImplementedError(
                        "non-numeric type is not supported for now"
                    )
            return self.new_dataframe(
                [df_or_series],
                shape=test_df.shape,
                dtypes=test_df.dtypes,
                index_value=parse_index(test_df.index, store_data=True),
                columns_value=parse_index(test_df.columns, store_data=True),
            )

    @classmethod
    def tile(cls, op):
        inp = op.input

        if len(inp.chunks) == 1:
            return cls._tile_one_chunk(op)

        if isinstance(inp, SERIES_TYPE):
            result = yield from cls._tile_series(op)
        else:
            result = yield from cls._tile_dataframe(op)
        return result

    @classmethod
    def _tile_one_chunk(cls, op):
        out = op.outputs[0]

        chunk_op = op.copy().reset_key()
        chunk_params = out.params.copy()
        chunk_params["index"] = (0,) * out.ndim
        out_chunk = chunk_op.new_chunk([op.input.chunks[0]], kws=[chunk_params])

        new_op = op.copy()
        params = out.params.copy()
        params["chunks"] = [out_chunk]
        params["nsplits"] = tuple((s,) for s in out.shape)
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def _tile_series(cls, op):
        series = Series(op.input)
        out = op.outputs[0]
        index = out.index_value.to_pandas()
        # ['count', 'mean', 'std', 'min', {percentiles}, 'max']
        names = index.tolist()

        values = [None] * 6
        for i, agg in enumerate(names[:4]):
            values[i] = mt.atleast_1d(getattr(series, agg)())
        values[-1] = mt.atleast_1d(getattr(series, names[-1])())
        values[4] = series.quantile(op.percentiles).to_tensor()

        t = mt.concatenate(values).rechunk(len(names))
        ret = Series(t, index=index, name=series.name)
        ret = yield from recursive_tile(ret)
        return [ret]

    @classmethod
    def _tile_dataframe(cls, op):
        df = DataFrame(op.input)
        out = op.outputs[0]
        dtypes = out.dtypes
        columns = dtypes.index.tolist()

        if df.chunk_shape[1] > 1:
            df = df.rechunk({1: df.shape[1]})

        # check dtypes if selected all fields
        # to reduce graph scale
        if df.dtypes.index.tolist() != columns:
            df = df[columns]

        # calculate percentiles
        percentiles = None
        if len(op.percentiles) > 0:
            if has_unknown_shape(*op.inputs):
                yield
            percentiles = yield from recursive_tile(df.quantile(op.percentiles))

        # perform aggregation together
        aggregation = yield from recursive_tile(
            df.agg(["count", "mean", "std", "min", "max"])
        )

        chunk_op = DataFrameDescribe(
            output_types=op.output_types,
            stage=OperandStage.agg,
            percentiles=op.percentiles,
        )
        chunk_params = out.params.copy()
        chunk_params["index"] = (0, 0)
        in_chunks = aggregation.chunks
        if percentiles is not None:
            in_chunks += percentiles.chunks
        out_chunk = chunk_op.new_chunk(in_chunks, kws=[chunk_params])

        new_op = op.copy()
        params = out.params.copy()
        params["chunks"] = [out_chunk]
        params["nsplits"] = tuple((s,) for s in out.shape)
        return new_op.new_tileables(op.inputs, kws=[params])

    @classmethod
    def execute(cls, ctx, op):
        out = op.outputs[0]
        if op.stage is None:  # 1 chunk
            df_or_series = ctx[op.input.key]

            ctx[out.key] = df_or_series.describe(
                percentiles=op.percentiles, include=op.include, exclude=op.exclude
            )
        else:
            assert op.stage == OperandStage.agg

            inputs = [ctx[inp.key] for inp in op.inputs]
            xdf = (
                pd
                if isinstance(inputs[0], (pd.DataFrame, pd.Series, pd.Index))
                or cudf is None
                else cudf
            )

            if len(inputs) == 1:
                df = inputs[0]
            else:
                assert len(inputs) > 1
                aggregations = inputs[0]
                percentiles = xdf.concat(inputs[1:], axis=0)
                df = xdf.concat(
                    [aggregations.iloc[:-1], percentiles, aggregations.iloc[-1:]],
                    axis=0,
                )
            # ['count', 'mean', 'std', 'min', {percentiles}, 'max']
            df.index = out.index_value.to_pandas()
            ctx[out.key] = df


def describe(df_or_series, percentiles=None, include=None, exclude=None):
    if percentiles is False:
        percentiles = []
    elif percentiles is None:
        percentiles = [0.25, 0.5, 0.75]
    else:
        percentiles = list(percentiles)
        if percentiles is not None:
            for p in percentiles:
                if p < 0 or p > 1:
                    raise ValueError(
                        "percentiles should all be in the interval [0, 1]. "
                        "Try [{0:.3f}] instead.".format(p / 100)
                    )
        # median should always be included
        if 0.5 not in percentiles:
            percentiles.append(0.5)
        percentiles = np.asarray(percentiles)

        # sort and check for duplicates
        unique_pcts = np.unique(percentiles)
        if len(unique_pcts) < len(percentiles):
            raise ValueError("percentiles cannot contain duplicates")
        percentiles = unique_pcts.tolist()

    op = DataFrameDescribe(percentiles=percentiles, include=include, exclude=exclude)
    return op(df_or_series)



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import namedtuple

import numpy as np
import pandas as pd

from .... import opcodes
from ....serialization.serializables import BoolField, Float64Field
from ..aggregation import BaseDataFrameExpandingAgg

_stage_info = namedtuple(
    "_stage_info",
    (
        "map_groups",
        "map_sources",
        "combine_sources",
        "combine_columns",
        "combine_funcs",
        "key_to_funcs",
        "valid_columns",
        "min_periods_func_name",
    ),
)

_cum_alpha_coeff_func = "_cum_alpha_coeff"
_cum_square_alpha_coeff_func = "_cum_square_alpha_coeff"


def _add_pred_results(
    pred_results,
    local_results,
    axis=0,
    alpha=None,
    order=1,
    alpha_ignore_na=False,
    pred_exponent=None,
    alpha_data=None,
):
    if pred_results[0].ndim == 1:
        df_filler = 0
    else:
        df_filler = pred_results[0].iloc[-1, :].dropna()
        df_filler[:] = 0

    new_locals = []
    combine_axis = pred_results[0].ndim - axis - 1
    weight = (1 - alpha) ** order
    pred_coeff = weight**pred_exponent
    for idx, (pred_result, local_result) in enumerate(zip(pred_results, local_results)):
        local_result.fillna(df_filler, inplace=True)
        pred_result = pred_result.mul(pred_coeff).sum(axis=axis)

        if alpha_ignore_na:
            pred_df = pred_result * weight ** alpha_data.notna().cumsum()
        else:
            weights = np.arange(1, len(local_result) + 1)
            if local_result.ndim == 2:
                weights_df = pd.DataFrame(
                    np.repeat(
                        weights.reshape((len(local_result), 1)),
                        len(local_result.columns),
                        axis=1,
                    ),
                    columns=local_result.columns,
                    index=local_result.index,
                )
            else:
                weights_df = pd.Series(weights, index=local_result.index)
            weights_df[alpha_data.isna()] = np.nan
            weights_df.ffill(inplace=True)
            weights_df.fillna(0, inplace=True)

            weights_df = weight**weights_df
            pred_df = weights_df.mul(pred_result, axis=combine_axis)

        new_locals.append(local_result.add(pred_df, axis=combine_axis))
    return new_locals


def _combine_mean(
    pred_results,
    local_results,
    axis=0,
    alpha=None,
    alpha_ignore_na=False,
    pred_exponent=None,
):
    if pred_results is None:
        return (local_results[0] / local_results[1]).ffill()

    alpha_data = local_results[1]
    local_results[0] = local_results[0].ffill()
    local_results[1] = alpha_data.ffill()

    local_sum_data, local_count_data = local_results

    if pred_results is not None:
        local_sum_data, local_count_data = _add_pred_results(
            pred_results,
            local_results,
            axis=axis,
            alpha=alpha,
            alpha_ignore_na=alpha_ignore_na,
            pred_exponent=pred_exponent,
            alpha_data=alpha_data,
        )
    return local_sum_data / local_count_data


def _combine_var(
    pred_results,
    local_results,
    axis=0,
    alpha=None,
    alpha_ignore_na=False,
    pred_exponent=None,
):
    local_results[0] = local_results[0].ffill()
    alpha_data = local_results[1]
    local_results[1] = alpha_data.ffill()

    local_results[2] = local_results[2].ffill()
    alpha2_data = local_results[3]
    local_results[3] = alpha2_data.ffill()

    (
        local_sum_data,
        local_count_data,
        local_sum_square,
        local_count2_data,
    ) = local_results
    if pred_results is None:
        return (local_sum_square - local_sum_data**2 / local_count_data) / (
            local_count_data - local_count2_data / local_count_data
        )

    pred_sum_data, pred_count_data, pred_sum_square, pred_count2_data = pred_results

    (local_count2_data,) = _add_pred_results(
        [pred_count2_data],
        [local_count2_data],
        axis=axis,
        alpha=alpha,
        order=2,
        alpha_ignore_na=alpha_ignore_na,
        pred_exponent=pred_exponent,
        alpha_data=alpha_data,
    )

    local_sum_square, local_sum_data, local_count_data = _add_pred_results(
        [pred_sum_square, pred_sum_data, pred_count_data],
        [local_sum_square, local_sum_data, local_count_data],
        axis=axis,
        alpha=alpha,
        alpha_ignore_na=alpha_ignore_na,
        pred_exponent=pred_exponent,
        alpha_data=alpha_data,
    )

    return (local_sum_square - local_sum_data**2 / local_count_data) / (
        local_count_data - local_count2_data / local_count_data
    )


def _combine_std(
    pred_results,
    local_results,
    axis=0,
    alpha=None,
    alpha_ignore_na=False,
    pred_exponent=None,
):
    return np.sqrt(
        _combine_var(
            pred_results,
            local_results,
            axis=axis,
            alpha=alpha,
            alpha_ignore_na=alpha_ignore_na,
            pred_exponent=pred_exponent,
        )
    )


def _combine_data_count(pred_results, local_results, axis=0, **__):
    if pred_results is None:
        return local_results[0]
    return local_results[0].add(
        pred_results[0].sum(), axis=pred_results[0].ndim - axis - 1
    )


class DataFrameEwmAgg(BaseDataFrameExpandingAgg):
    _op_type_ = opcodes.EWM_AGG

    _alpha = Float64Field("alpha")
    _adjust = BoolField("adjust")
    _alpha_ignore_na = BoolField("alpha_ignore_na")

    _validate_columns = BoolField("_validate_columns")

    _exec_cache = dict()

    def __init__(
        self, alpha=None, adjust=None, alpha_ignore_na=None, validate_columns=None, **kw
    ):
        super().__init__(
            _alpha=alpha,
            _adjust=adjust,
            _alpha_ignore_na=alpha_ignore_na,
            _validate_columns=validate_columns,
            **kw
        )

    @property
    def alpha(self) -> float:
        return self._alpha

    @property
    def adjust(self) -> bool:
        return self._adjust

    @property
    def alpha_ignore_na(self) -> bool:
        return self._alpha_ignore_na

    @property
    def validate_columns(self) -> bool:
        return self._validate_columns

    @classmethod
    def _get_stage_functions(cls, op: "DataFrameEwmAgg", func):
        if func == "_data_count":
            return ["_data_count"], _combine_data_count
        elif func == "mean":
            return ["cumsum", _cum_alpha_coeff_func], _combine_mean
        elif func in {"var", "std"}:
            return (
                [
                    "cumsum",
                    _cum_alpha_coeff_func,
                    "cumsum2",
                    _cum_square_alpha_coeff_func,
                ],
                _combine_var if func == "var" else _combine_std,
            )
        else:  # pragma: no cover
            raise NotImplementedError

    @classmethod
    def _calc_data_alphas(cls, op: "DataFrameEwmAgg", in_data, order):
        exec_cache = cls._exec_cache[op.key]
        cache_key = ("_calc_data_alphas", order, id(in_data))
        try:
            return exec_cache[cache_key]
        except KeyError:
            pass

        cum_df = in_data.copy()
        cum_df[cum_df.notna()] = 1
        if not op.alpha_ignore_na:
            cum_df.ffill(inplace=True)
        cum_df = cum_df.cumsum(axis=op.axis) - 1
        if not op.alpha_ignore_na:
            cum_df[in_data.isna()] = np.nan

        result = exec_cache[cache_key] = (1 - op.alpha) ** (order * cum_df)
        return result

    @classmethod
    def _execute_cum_alpha_coeff(
        cls, op: "DataFrameEwmAgg", in_data, order, final=True
    ):
        exec_cache = cls._exec_cache[op.key]
        cache_key = ("cum_alpha_coeff", order, id(in_data))
        summary = None

        try:
            result = exec_cache[cache_key]
        except KeyError:
            alphas = cls._calc_data_alphas(op, in_data, order)
            result = alphas.cumsum()
            exec_cache[cache_key] = result

        if final:
            if op.output_agg:
                summary = result.ffill()[-1:]
        return result, summary

    @classmethod
    def _execute_cumsum(cls, op: "DataFrameEwmAgg", in_data):
        exec_cache = cls._exec_cache[op.key]
        cache_key = ("cumsum", id(in_data))
        summary = None

        try:
            result = exec_cache[cache_key]
        except KeyError:
            min_periods = 1 if op.min_periods > 0 else 0

            try:
                data = in_data.ewm(
                    alpha=op.alpha,
                    ignore_na=op.alpha_ignore_na,
                    adjust=op.adjust,
                    min_periods=min_periods,
                ).mean()
            except ValueError:
                in_data = in_data.copy()
                data = in_data.ewm(
                    alpha=op.alpha,
                    ignore_na=op.alpha_ignore_na,
                    adjust=op.adjust,
                    min_periods=min_periods,
                ).mean()

            alpha_sum, _ = op._execute_cum_alpha_coeff(op, in_data, 1, final=False)
            result = exec_cache[cache_key] = data * alpha_sum

        if op.output_agg:
            summary = result.ffill()[-1:]
        return result, summary

    @classmethod
    def _execute_cumsum2(cls, op: "DataFrameEwmAgg", in_data):
        summary = None
        min_periods = 1 if op.min_periods > 0 else 0

        try:
            data = in_data.ewm(
                alpha=op.alpha,
                ignore_na=op.alpha_ignore_na,
                adjust=op.adjust,
                min_periods=min_periods,
            ).var(bias=True)
        except ValueError:
            in_data = in_data.copy()
            data = in_data.ewm(
                alpha=op.alpha,
                ignore_na=op.alpha_ignore_na,
                adjust=op.adjust,
                min_periods=min_periods,
            ).var(bias=True)

        alpha_sum, _ = op._execute_cum_alpha_coeff(op, in_data, 1)
        cumsum, _ = op._execute_cumsum(op, in_data)
        result = alpha_sum * data + cumsum**2 / alpha_sum

        if op.output_agg:
            summary = result.ffill()[-1:]

        return result, summary

    @classmethod
    def _execute_map_function(cls, op: "DataFrameEwmAgg", func, in_data):
        in_data = in_data._get_numeric_data()

        summary = None
        min_periods = 1 if op.min_periods > 0 else 0
        if func == "_data_count":
            result = in_data.expanding(min_periods=min_periods).count()
        elif func in (_cum_alpha_coeff_func, _cum_square_alpha_coeff_func):
            order = 1 if func == _cum_alpha_coeff_func else 2
            result, summary = cls._execute_cum_alpha_coeff(op, in_data, order)
        elif func == "cumsum":
            result, summary = cls._execute_cumsum(op, in_data)
        elif func == "cumsum2":
            result, summary = cls._execute_cumsum2(op, in_data)
        else:  # pragma: no cover
            raise ValueError("Map function %s not supported")

        if op.output_agg:
            summary = summary if summary is not None else result.iloc[-1:]
        else:
            summary = None
        return result, summary

    @classmethod
    def _execute_map(cls, ctx, op: "DataFrameEwmAgg"):
        try:
            cls._exec_cache[op.key] = dict()

            super()._execute_map(ctx, op)
            if op.output_agg:
                in_data = ctx[op.inputs[0].key]
                summaries = list(ctx[op.outputs[1].key])

                if op.alpha_ignore_na:
                    in_count = in_data.count()
                    if not isinstance(in_count, pd.Series):
                        in_count = pd.Series([in_count])
                    summary = in_count
                    if in_data.ndim == 2:
                        summary = in_count.to_frame().T
                    summary.index = summaries[-1].index
                else:
                    remain_counts = in_data.notna()[::-1].to_numpy().argmax(axis=0)
                    if in_data.ndim > 1:
                        remain_counts = remain_counts.reshape((1, len(in_data.columns)))
                        summary = pd.DataFrame(
                            remain_counts,
                            columns=in_data.columns,
                            index=summaries[-1].index,
                        )
                    else:
                        summary = pd.Series(remain_counts, index=summaries[-1].index)
                summaries.insert(-1, summary)

                ctx[op.outputs[1].key] = tuple(summaries)
        finally:
            cls._exec_cache.pop(op.key, None)

    @classmethod
    def _execute_combine_function(
        cls, op: "DataFrameEwmAgg", func, prev_inputs, local_inputs, func_cols
    ):
        exec_cache = cls._exec_cache[op.key]
        pred_exponent = exec_cache.get("pred_exponent")
        if func_cols and pred_exponent is not None:
            pred_exponent = (
                pred_exponent[func_cols] if pred_exponent is not None else None
            )
        return func(
            prev_inputs,
            local_inputs,
            axis=op.axis,
            alpha=op.alpha,
            alpha_ignore_na=op.alpha_ignore_na,
            pred_exponent=pred_exponent,
        )

    @classmethod
    def _execute_combine(cls, ctx, op: "DataFrameEwmAgg"):
        try:
            cls._exec_cache[op.key] = dict()

            if len(op.inputs) != 1:
                pred_data = ctx[op.inputs[1].key]

                if op.alpha_ignore_na:
                    pred_exponent = (
                        pred_data[-2].shift(-1)[::-1].cumsum()[::-1].fillna(0)
                    )
                else:
                    succ_counts = pred_data[-1].shift(-1)
                    succ_counts.iloc[-1] = 0
                    pred_exponent = pred_data[-2].add(
                        succ_counts[::-1].cumsum()[::-1], axis=op.axis
                    )

                cls._exec_cache[op.key]["pred_exponent"] = pred_exponent

            super()._execute_combine(ctx, op)
        finally:
            cls._exec_cache.pop(op.key, None)

    @classmethod
    def _execute_raw_function(cls, op: "DataFrameEwmAgg", in_data):
        for _ in range(2):
            ewm = in_data.ewm(
                alpha=op.alpha,
                min_periods=op.min_periods,
                adjust=op.adjust,
                ignore_na=op.alpha_ignore_na,
            )
            try:
                val = ewm.agg(op.func)
                if (
                    in_data.ndim == 2
                    and op.validate_columns
                    and len(val.columns) != len(op.outputs[0].columns_value.to_pandas())
                ):
                    raise ValueError("Columns not consistent")
                return val
            except ValueError:
                in_data = in_data.copy()
        else:  # pragma: no cover
            raise ValueError



# Copyright 1999-2021 Alibaba Group Holding Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .... import oscar as mo
from ...core import AbstractService
from .core import MutableObjectManagerActor


class MutableObjectManagerSupervisorService(AbstractService):
    async def start(self):
        pass

    async def stop(self):
        pass

    async def create_session(self, session_id: str):
        await mo.create_actor(
            MutableObjectManagerActor,
            session_id,
            address=self._address,
            uid=MutableObjectManagerActor.gen_uid(session_id),
        )

    async def destroy_session(self, session_id: str):
        await mo.destroy_actor(
            mo.create_actor_ref(
                uid=MutableObjectManagerActor.gen_uid(session_id), address=self._address
            )
        )
